"""A collection of little command line utilities for inspecting mass spectrum data."""
import io
import logging
import os
import math
import csv
import sys
from collections import Counter
from typing import List

import click
import six

import numpy as np

import ms_deisotope
from ms_deisotope.data_source.scan.scan import Scan

from ms_deisotope.task.log_utils import init_logging

from ms_deisotope.feature_map import quick_index
from ms_deisotope.feature_map import scan_interval_tree

from ms_deisotope.clustering.scan_clustering import (
    iterative_clustering, ScanClusterWriter, ScanClusterReader, _DynamicallyLoadingResolver)

from ms_deisotope.qc.isolation import isolation_window_valid, is_isolation_window_empty
from ms_deisotope.qc.signature import TMTReporterExtractor

from ms_deisotope.data_source import (
    _compression, ScanProxyContext, MSFileLoader)
from ms_deisotope.data_source.scan import RandomAccessScanSource
from ms_deisotope.data_source.metadata.file_information import SourceFile

from ms_deisotope.output import ProcessedMzMLDeserializer

from ms_deisotope.tools import conversion, draw, maintenance
from ms_deisotope.tools.utils import processes_option, is_debug_mode, progress, register_debug_hook, spinner


CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])

logger = logging.getLogger('ms_deisotope')


@click.group(context_settings=CONTEXT_SETTINGS)
def cli():
    """A collection of utilities for inspecting and manipulating mass spectrometry data."""
    init_logging()


@cli.command("describe", short_help=("Produce a minimal textual description"
                                     " of a mass spectrometry data file"))
@click.argument('path', type=click.Path(exists=True))
@click.option("-d", "--diagnostics", is_flag=True,
              help="Run more diagnostics, greatly increasing runtime but producing additional informatoin")
def describe(path, diagnostics=False):
    """Produces a minimal textual description of a mass spectrometry data file."""
    click.echo("Describing \"%s\"" % (path,))
    try:
        sf = SourceFile.from_path(path)
    except IOError as err:
        raise click.ClickException("Could not open file \"%s\":\n%s" % (path, err))

    if sf.file_format is None:
        raise click.ClickException("\"%s\" doesn't appear to be a mass spectrometry data file" % (path, ))
    click.echo("File Format: %s" % (sf.file_format, ))
    click.echo("ID Format: %s" % (sf.id_format, ))
    reader = MSFileLoader(path)
    if isinstance(reader, RandomAccessScanSource):
        click.echo("Format Supports Random Access: True")
        first_scan = reader[0]
        last_scan = reader[-1]
        click.echo("First Scan: %s at %0.3f minutes" % (first_scan.id, first_scan.scan_time))
        click.echo("Last Scan: %s at %0.3f minutes" % (last_scan.id, last_scan.scan_time))
    else:
        click.echo("Format Supports Random Access: False")
    try:
        finfo = reader.file_description()
        click.echo("Contents:")
        for key in finfo.contents:
            click.echo("    %s" % (key, ))
    except AttributeError:
        pass
    index_file_name = quick_index.ExtendedScanIndex.index_file_name(path)
    # Extra introspection if the extended index is available
    if os.path.exists(index_file_name):
        with open(index_file_name, 'rt') as fh:
            index = quick_index.ExtendedScanIndex.deserialize(fh)
        ms1_scans = len(index.ms1_ids)
        msn_scans = len(index.msn_ids)
        click.echo("MS1 Scans: %d" % (ms1_scans, ))
        click.echo("MSn Scans: %d" % (msn_scans, ))
        n_defaulted = 0
        n_orphan = 0

        charges = Counter()
        first_msn = float('inf')
        last_msn = 0
        for scan_info in index.msn_ids.values():
            n_defaulted += scan_info.get('defaulted', False)
            n_orphan += scan_info.get('orphan', False)
            charges[scan_info['charge']] += 1
            rt = scan_info['scan_time']
            if rt < first_msn:
                first_msn = rt
            if rt > last_msn:
                last_msn = rt
        click.echo("First MSn Scan: %0.3f minutes" % (first_msn,))
        click.echo("Last MSn Scan: %0.3f minutes" % (last_msn,))
        for charge, count in sorted(charges.items()):
            if not isinstance(charge, int):
                continue
            click.echo("Precursors with Charge State %d: %d" % (charge, count))
        if n_defaulted > 0:
            click.echo("Defaulted MSn Scans: %d" % (n_defaulted,))
        if n_orphan > 0:
            click.echo("Orphan MSn Scans: %d" % (n_orphan,))
    if diagnostics:
        reader.reset()
        scan_ids_with_invalid_isolation_window = []
        scan_ids_with_empty_isolation_window = []
        activation_counter = Counter()
        n_ms1 = 0
        n_msn = 0
        for precursor, products in reader:
            if precursor is not None:
                n_ms1 += 1
            for product in products:
                n_msn += 1
                if not isolation_window_valid(product):
                    scan_ids_with_invalid_isolation_window.append((precursor.id, product.id))
                if is_isolation_window_empty(product):
                    scan_ids_with_empty_isolation_window.append(
                        (precursor.id, product.id))
                activation_counter[product.activation] += 1

        click.echo("MS1 Spectra: %d" % (n_ms1, ))
        click.echo("MSn Spectra: %d" % (n_msn, ))
        click.echo("Invalid Isolation Windows: %d" % (
            len(scan_ids_with_invalid_isolation_window), ))
        click.echo("Empty Isolation Windows: %d" % (
            len(scan_ids_with_empty_isolation_window), ))

        click.echo("Activation Methods")
        for activation, count in activation_counter.items():
            if not activation.is_multiple_dissociation():
                click.echo("\t{method}:{energy} = {count}".format(
                    method=activation.method, energy=activation.energy, count=count))
            else:
                click.echo("\t{method}:{energy} = {count}".format(
                    method=','.join(map(str, activation.method)),
                    energy=','.join(map(str, activation.energies)), count=count))


@cli.command("byte-index", short_help='Build an external byte offset index for a mass spectrometry data file')
@click.argument('paths', type=click.Path(exists=True), nargs=-1)
def byte_index(paths):
    """Build an external byte offset index for a MS data file, saving time when opening the file with indexing enabled.

    Supported Formats: mzML, mzXML, MGF
    """
    for path in paths:
        click.echo("Indexing %s" % (path, ))
        reader = ms_deisotope.MSFileLoader(path, use_index=False)
        try:
            fn = reader.prebuild_byte_offset_file
        except AttributeError:
            click.echo("\"%s\" does not support pre-indexing byte offsets" % (path,))
            return
        fn(path)


@cli.command("metadata-index", short_help='Build an external scan metadata index for a mass spectrometry data file')
@click.argument('paths', type=click.Path(exists=True), nargs=-1)
@click.option("-D", '--deconvoluted', is_flag=True, help='Whether or not to assume this file has been '
              'processed specifically by ms_deisotope\'s deconvolution algorithm')
@processes_option
def metadata_index(paths, processes=4, deconvoluted=False):
    """Build an external scan metadata index for a mass spectrometry data file.

    This extended index is saved in a separate JSON file that can be loaded with
    :class:`~.ExtendedScanIndex`. It includes the scan time of all scans, the precursor
    mass of MSn scans, as well as the relationships between precursor and product ion
    scans, as well as other details. See :class:`~.ExtendedScanIndex` for more information
    """
    for path in paths:
        click.echo("Indexing %s" % (path, ))
        if deconvoluted:
            reader = ProcessedMzMLDeserializer(path, use_extended_index=False)
        else:
            reader = MSFileLoader(path)
        try:
            fn = reader.prebuild_byte_offset_file
            if not reader.source._check_has_byte_offset_file():
                fn(path)
        except AttributeError:
            pass
        if processes > 1:
            progbar = progress(label='Building Index', length=100)
            acc = [0]

            def update_bar(x):
                """Progress Bar update callback for :func:`~.quick_index.index`."""
                x = int(x * 100)
                x -= acc[0]  # pylint: disable=cell-var-from-loop
                progbar.update(x)  # pylint: disable=cell-var-from-loop
                acc[0] += x  # pylint: disable=cell-var-from-loop

            with progbar:
                update_bar(0.0)
                index, _ = quick_index.index(
                    reader, processes, progress_indicator=update_bar)
        else:
            index = quick_index.ExtendedScanIndex()
            reader.reset()
            try:
                n = len(reader)
                progbar = progress(label='Building Index', length=n)
            except TypeError:
                progbar = spinner(title="Building Index")
            with progbar:
                for bunch in reader.make_iterator(grouped=True):
                    i = 0
                    i += bunch.precursor is not None
                    i += len(bunch.products)
                    index.add_scan_bunch(bunch)
                    progbar.update(i)

        name = path
        index_file_name = index.index_file_name(name)
        with open(index_file_name, 'w') as fh:
            index.serialize(fh)


def _partial_ms_file_iterator(reader, start, end):
    for scan_bunch in reader.start_from_scan(index=start):
        if scan_bunch.precursor.index > end:
            break
        yield scan_bunch


class _MSMSIntervalTask(object):
    def __init__(self, time_radius, mz_lower, mz_higher):
        self.time_radius = time_radius
        self.mz_lower = mz_lower
        self.mz_higher = mz_higher

    def __call__(self, payload):
        reader, start, end = payload
        iterator = _partial_ms_file_iterator(reader, start, end)
        intervals = scan_interval_tree.extract_intervals(
            iterator, time_radius=self.time_radius,
            mz_lower=self.mz_lower, mz_higher=self.mz_higher)
        return intervals


@cli.command("msms-intervals", short_help=(
    'Build an interval tree over precursor isolation events in time and m/z space'))
@click.argument('paths', type=click.Path(exists=True), nargs=-1)
@click.option("-o", "--output", type=click.Path(writable=True, file_okay=True, dir_okay=False), required=False)
@processes_option
def msms_intervals(paths, processes=4, time_radius=5, mz_lower=2., mz_higher=3., output=None):
    """Construct an interval tree spanning time and m/z domains where MSn spectra were acquired in the LC-MS map.

    The interval tree is serialized to JSON.
    """
    interval_extraction = _MSMSIntervalTask(time_radius, mz_lower, mz_higher)
    interval_set = []
    total_work_items = len(paths) * processes * 4

    def _run():
        for path in paths:
            reader = MSFileLoader(path)
            chunk_out_of_order = quick_index.run_task_in_chunks(
                reader, processes, processes * 4, task=interval_extraction)
            for chunk in chunk_out_of_order:
                interval_set.extend(chunk)
                yield 0
    work_iterator = _run()
    with progress(work_iterator, length=total_work_items, label='Extracting Intervals',
                  color=True, fill_char=click.style('-', 'green')) as g:
        for _ in g:
            pass
    tree = scan_interval_tree.ScanIntervalTree(scan_interval_tree.make_rt_tree(interval_set))
    if output is not None:
        with open(output, 'wt') as fh:
            tree.serialize(fh)
    else:
        stream = click.get_text_stream('stdout')
        tree.serialize(stream)
        stream.flush()


def _ensure_metadata_index(path):
    reader = ms_deisotope.MSFileLoader(path)
    name = path
    index_file_name = quick_index.ExtendedScanIndex.index_file_name(name)
    if not os.path.exists(index_file_name):
        click.secho("Building Index For %s" % (path, ), fg='yellow', err=True)
        index = quick_index.ExtendedScanIndex()
        reader.reset()
        try:
            n = len(reader)
            progbar = click.progressbar(
                label='Building Index', length=n, color=True,
                fill_char=click.style('-', 'green'))
        except TypeError:
            progbar = spinner(title="Building Index")
        with progbar:
            for bunch in reader.make_iterator(grouped=True):
                i = 0
                i += bunch.precursor is not None
                i += len(bunch.products)
                index.add_scan_bunch(bunch)
                progbar.update(i)
        reader.reset()
        with open(index_file_name, 'wt') as fh:
            index.serialize(fh)
    else:
        with open(index_file_name, 'rt') as fh:
            index = quick_index.ExtendedScanIndex.deserialize(fh)
    return reader, index


@cli.command("charge-states", short_help='Count the different precursor charge states in a mass spectrometry data file')
@click.argument("path", type=click.Path(exists=True))
def charge_states(path):
    """Count the different precursor charge states in a mass spectrometry data file.

    This command will construct a metadata index if it is not found.
    """
    _, index = _ensure_metadata_index(path)

    charges = Counter()
    for _, msn_info in index.msn_ids.items():
        charges[msn_info.charge] += 1

    def sort_key(k):
        if isinstance(k, int):
            return abs(k)
        else:
            return 0

    for charge in sorted(charges, key=sort_key):
        click.echo("%s: %d" % (charge, charges[charge]))


def _binsearch(array, x):
    n = len(array)
    lo = 0
    hi = n

    while hi != lo:
        mid = (hi + lo) // 2
        y = array[mid][0]
        err = y - x
        if hi - lo == 1:
            return mid
        elif err > 0:
            hi = mid
        else:
            lo = mid
    return


@cli.command("precursor-clustering", short_help='Cluster precursor masses in a mass spectrometry data file')
@click.argument("path", type=click.Path(exists=True))
def precursor_clustering(path, grouping_error=2e-5):
    """Cluster precursor masses in a mass spectrometry data file.

    This command will construct a metadata index if it is not found.
    """
    _, index = _ensure_metadata_index(path)
    points = []
    for _, msn_info in index.msn_ids.items():
        points.append((msn_info.neutral_mass, msn_info.intensity))
    points.sort(key=lambda x: x[1], reverse=1)
    centroids = []
    if not points:
        click.secho("No MS/MS detected", fg='yellow', err=True)
        return

    for point in points:
        if not centroids:
            centroids.append((point[0], [point]))
            continue
        i = _binsearch(centroids, point[0])
        centroid = centroids[i]
        err = (centroid[0] - point[0]) / point[0]
        if abs(err) < grouping_error:
            centroid[1].append(point)
        else:
            if err < 0:
                i += 1
            centroids.insert(i, (point[0], [point]))
    acc = 0
    nt = 0
    for centroid, obs in centroids:
        n = len(obs)
        if n == 1:
            continue
        mean = sum(p[0] for p in obs) / n
        var = sum([(p[0] - mean) ** 2 for p in obs]) / (n - 1)
        acc += (var * (n - 1))
        nt += (n - 1)
    for centroid, obs in centroids:
        click.echo("%f: %d" % (centroid, len(obs)))
    click.echo("MS/MS Precursor Mass Std. Dev.: %f Da" % (math.sqrt(acc / nt),))


@cli.command("spectrum-clustering", short_help=("Cluster MSn spectra in a mass spectrometry data file using"
                                                " cosine similarity"))
@click.argument('paths', type=click.Path(exists=True), nargs=-1)
@click.option("-m", "--precursor-error-tolerance", type=float, default=1e-5)
@click.option("-t", "--similarity-threshold", "similarity_thresholds", multiple=True, type=float)
@click.option("-o", "--output", "output_path", type=click.Path(writable=True, file_okay=True, dir_okay=False),
              required=False)
@click.option("-c", "--cache-size", type=int, default=2**10, help=(
    "The number of scans to cache in memory when not using --in-memory. If you are clustering multiple "
    "files, this number will be apply to each file separately."))
@click.option("-M", "--in-memory", is_flag=True, default=False, help=(
    "Whether to load the entire dataset into memory for better performance"))
@click.option("-D", "--deconvoluted", is_flag=True, default=False, help=(
    "Whether to assume the spectra are deconvoluted or not"))
def spectrum_clustering(paths, precursor_error_tolerance=1e-5, similarity_thresholds=None, output_path=None,
                        in_memory=False, deconvoluted=False, cache_size=2**10):
    """Cluster spectra by precursor mass and cosine similarity.

    Spectrum clusters are written out to a text file recording
    cluster precursor mass, within-cluster similarity, and the
    source file and scan ID for each cluster member.
    """
    if not similarity_thresholds:
        similarity_thresholds = [0.1, 0.4, 0.7]
    else:
        similarity_thresholds = sorted(similarity_thresholds)
    if output_path is None:
        output_path = "-"
    msn_scans = []
    n_spectra = 0

    with progress(paths, label="Indexing", item_show_func=lambda x: str(x) if x else '',
                  color=True, fill_char=click.style('-', 'cyan')) as progbar:
        key_seqs = []
        for path in progbar:
            if deconvoluted:
                reader = ProcessedMzMLDeserializer(path)
                reader.parse_envelopes = False
                index = reader.extended_index
            else:
                reader, index = _ensure_metadata_index(path)
            key_seqs.append((reader, index))
            n_spectra += len(index.msn_ids)

    with progress(label="Loading Spectra", length=n_spectra,
                  item_show_func=lambda x: str(x) if x else '', color=True,
                  fill_char=click.style('-', 'green')) as progbar:
        for reader, index in key_seqs:
            if not in_memory:
                if not reader.has_fast_random_access:
                    click.secho(
                        "%s does not have fast random access, scan fetching may be slow!" % (
                            reader, ), fg='yellow')
                proxy_context = ScanProxyContext(reader, cache_size=cache_size)
                pinfo_map = {
                    pinfo.product_scan_id: pinfo for pinfo in
                    index.get_precursor_information()
                }
                for i in index.msn_ids:
                    progbar.current_item = i
                    progbar.update(1)
                    scan = proxy_context(i)
                    scan.precursor_information = pinfo_map[i]
                    msn_scans.append(scan)
            else:
                if reader.has_fast_random_access:
                    # We have fast random access so we can just loop over the index and pull out
                    # the MSn scans directly without completely traversing the file.
                    for i in index.msn_ids:
                        progbar.current_item = i
                        progbar.update(1)
                        scan = reader.get_scan_by_id(i)
                        if scan.peak_set is None and not deconvoluted:
                            scan = scan.pick_peaks().pack(bind=True)
                        msn_scans.append(scan)
                else:
                    # If we don't  have fast random access, it's better just to loop over the file,
                    # and absorb the cost of parsing the MS1 scans
                    reader.reset()
                    reader.make_iterator(grouped=False)
                    for scan in reader:
                        if scan.ms_level != 1:
                            progbar.current_item = scan.id
                            progbar.update(1)
                            if scan.peak_set is None and not deconvoluted:
                                scan = scan.pick_peaks().pack(bind=True)
                            msn_scans.append(scan)
                # Dispose of the state that is no longer required.
                reader.reset()
                index.clear()


    click.echo("Begin Clustering", err=True)
    clusters = iterative_clustering(
        msn_scans, precursor_error_tolerance, similarity_thresholds)
    click.echo("Clusering Finished", err=True)
    by_size = Counter()
    for cluster in clusters:
        by_size[len(cluster)] += 1
    click.echo("Clusters: {:d}".format(len(clusters)), err=True)
    for key, value in sorted(by_size.items()):
        click.echo("Size {:d}: {:d}".format(key, value), err=True)
    with click.open_file(output_path, mode='w') as outfh:
        writer = ScanClusterWriter(outfh)
        for cluster in clusters:
            writer.save(cluster)


@cli.command('cluster-statistics')
@click.argument('path', type=click.Path(readable=True, dir_okay=False))
def cluster_evaluation(path):
    """Calculate some statistics describing the clustered mass spectra described in the specified cluster file."""
    with click.open_file(path) as fh:
        dirname = os.path.dirname(path)
        reader = ScanClusterReader(fh, _DynamicallyLoadingResolver(dirname))
        clusters = list(reader)
        by_size = Counter()
        for cluster in clusters:
            by_size[len(cluster)] += 1
        click.echo("Clusters: {:d}".format(len(clusters)))
        for key, value in sorted(by_size.items()):
            click.echo("Size {:d}: {:d}".format(key, value))


@cli.command('ms1-spectrum-diagnostics')
@click.argument('path', type=click.Path(exists=True, readable=True))
@click.option("-o", "--output-path", type=click.Path(writable=True))
def ms1_spectrum_diagnostics(path, output_path=None):
    """Collect diagnostic information from MS1 spectra."""
    reader = ms_deisotope.MSFileLoader(path)

    reader.make_iterator(grouped=True)

    ms1_metric_names = [
        'scan_id', 'scan_index', 'scan_time', 'duty_cycle', 'tic',
        'base_peak_mz', 'base_peak_intensity', 'data_point_count',
        'injection_time', 'n_ms2_scans'
    ]
    ms1_metrics = []
    products = None
    last_ms1 = None
    prog = progress(length=len(reader), label='Processing Scans',
                    file=sys.stderr, item_show_func=lambda x: x.id if x else '',
                    color=True, fill_char=click.style('-', 'green'))
    with prog:
        for precursor, products in reader:
            ms1_time = precursor.scan_time
            if last_ms1 is not None:
                duty_cycle = ms1_time - last_ms1
                ms1_metrics[-1]['duty_cycle'] = duty_cycle
            last_ms1 = ms1_time
            bp = precursor.base_peak()
            acquisition_info = precursor.acquisition_information
            if acquisition_info:
                scan_event = acquisition_info[0]
                inj = scan_event.injection_time
            else:
                inj = np.nan
            ms1_record = {
                "scan_id": precursor.id,
                "scan_index": precursor.index,
                "scan_time": precursor.scan_time,
                "duty_cycle": np.nan,
                "tic": precursor.tic(),
                "base_peak_mz": bp.mz,
                "base_peak_intensity": bp.intensity,
                "data_point_count": precursor.arrays.mz.size,
                "injection_time": inj,
                "n_ms2_scans": len([p for p in products if p.ms_level == 2])
            }
            ms1_metrics.append(ms1_record)
            prog.current_item = precursor
            prog.update(1 + len(products))

    if last_ms1 is not None:
        if products:
            last_time = max([p.scan_time for p in products])
            duty_cycle = last_time - last_ms1
            ms1_metrics[-1]['duty_cycle'] = duty_cycle


    if output_path is None:
        outfh = click.open_file("-", mode='wb')
    else:
        outfh = io.open(output_path, mode='wb')
    if six.PY3:
        stream = io.TextIOWrapper(outfh, encoding='utf8', newline='')
    else:
        stream = outfh
    writer = csv.DictWriter(stream, fieldnames=ms1_metric_names)
    writer.writeheader()
    writer.writerows(ms1_metrics)
    stream.flush()


@cli.command("extract-reporter-ions", short_help="Extract reporter ion channels from an MS file to a CSV")
@click.argument('path', type=click.Path(exists=True, readable=True))
@click.option("-o", "--output-path", type=click.Path(writable=True), default='-',
              help="The path to write output to. Defaults to STDOUT")
@click.option("-r", "--reagent", default='tmt11', type=click.Choice(
    sorted(TMTReporterExtractor.TMT_REAGENTS)),
    required=False, help="The isobaric quantification tag reagent used")
@click.option("-m", "--error-tolerance", type=float, default=1e-5,
              help="The mass accuracy error tolerance to use when matching reporter ions")
def extract_reporter_ions(path, output_path=None, reagent='tmt11', error_tolerance=1e-5):
    """Extract the reporter ions from MSn spectra."""
    reader = ms_deisotope.MSFileLoader(path)

    if error_tolerance > 1e-3:
        logger.warn(
            f"Error tolerance {error_tolerance} looks like it is not in units of PPM, multiplying by 1e-6")
        error_tolerance *= 1e-6

    reader.make_iterator(grouped=True)

    extractor = TMTReporterExtractor(reagent=reagent)
    channels = [t.name for t in extractor.signature_ions]
    columns = ['scan_id', 'scan_time', 'precursor_mz', 'precursor_charge',
               'precursor_intensity', 'tic', 'precursor_purity'] + channels

    out_file = click.open_file(output_path, mode='wb')
    out_file_wrapper = io.TextIOWrapper(out_file, encoding='utf8', newline='')
    writer = csv.DictWriter(out_file_wrapper, columns)
    writer.writeheader()

    products: List[Scan]
    for _precursor, products in reader:
        for product in products:
            pinfo = product.precursor_information
            if product.is_profile:
                product.pick_peaks()
            row = {
                "scan_id": product.id,
                "scan_time": product.scan_time,
                "tic": product.tic(),
                "precursor_purity": product.annotations.get("precursor purity")
            }
            if pinfo is not None:
                row['precursor_mz'] = pinfo.mz
                row['precursor_charge'] = pinfo.charge if isinstance(pinfo.charge, int) else None
                row['precursor_intensity'] = pinfo.intensity

            row.update(extractor.extract(
                product, error_tolerance=error_tolerance))
            writer.writerow(row)



if _compression.has_idzip:
    @cli.command("idzip", short_help='Compress a file with idzip, a gzip-compatible format with random access support')
    @click.argument('path', type=click.Path(allow_dash=True, readable=True))
    @click.option("-o", "--output", type=click.Path(writable=True, file_okay=True, dir_okay=False),
                  required=False, help="The output stream to write to. Defaults to STDOUT if omitted.")
    def idzip_compression(path, output):
        """Compress a file using  idzip, a gzip-compatible format with random access support."""
        if output is None:
            output = '-'
        with click.open_file(output, mode='wb') as outfh:
            writer = _compression.GzipFile(fileobj=outfh, mode='wb')
            with click.open_file(path, 'rb') as infh:
                try:
                    infh_wrap = io.BufferedReader(infh)
                    header = infh_wrap.peek(2)
                    if _compression.starts_with_gz_magic(header):
                        click.echo("Detected gzip input file", err=True)
                        infh_wrap = _compression.GzipFile(fileobj=infh_wrap)
                except AttributeError:
                    infh_wrap = infh
                buffer_size = _compression.WRITE_BUFFER_SIZE
                chunk = infh_wrap.read(buffer_size)
                while chunk:
                    writer.write(chunk)
                    chunk = infh_wrap.read(buffer_size)
            writer.close()



def _mount_group(group):
    try:
        for name, command in group.commands.items():
            cli.add_command(command, name)
    except Exception as e:
        click.secho("%r occurred while loading additional tools from %r" % (e, group), err=True, fg='yellow')


_mount_group(conversion.ms_conversion)

cli.add_command(draw.draw)
cli.add_command(maintenance.maintenance)
main = cli.main

if is_debug_mode():
    register_debug_hook()

if __name__ == '__main__':
    click.secho("Running Debug Mode", fg='yellow')
    register_debug_hook()
    main()



"""
Writing mzML
------------

Using the :mod:`psims` library, :mod:`ms_deisotope.output.mzml` can write an mzML
file with all associated metadata, including deconvoluted peak arrays, chromatograms,
and data transformations. The :class:`~.MzMLSerializer` class handles all facets of
this process.

This module also contains a specialized version of :class:`~.MzMLLoader`,
:class:`~.ProcessedMzMLLoader`, which can directly reconstruct each
deconvoluted peak list and provides fast access to an extended index of
metadata that :class:`~.MzMLSerializer` writes to an external file.


.. code:: python

    import ms_deisotope
    from ms_deisotope.test.common import datafile
    from ms_deisotope.output.mzml import MzMLSerializer

    reader = ms_deisotope.MSFileLoader(datafile("small.mzML"))
    with open("small.deconvoluted.mzML", 'wb') as fh:
        writer = MzMLSerializer(fh, n_spectra=len(reader))

        writer.copy_metadata_from(reader)
        for bunch in reader:
            bunch.precursor.pick_peaks()
            bunch.precursor.deconvolute()
            for product in bunch.products:
                product.pick_peaks()
                product.deconvolute()
            writer.save(bunch)

        writer.close()

"""
import hashlib
import array
import io
import os
import warnings

from contextlib import contextmanager
from typing import Any, ContextManager, Dict, List, Tuple, Union, Optional, OrderedDict

from ms_deisotope.data_source.metadata.instrument_components import InstrumentInformation, InstrumentModel
from ms_deisotope.data_source.metadata.activation import ActivationInformation
from ms_deisotope.data_source.metadata.scan_traits import IsolationWindow
from ms_deisotope.data_source.metadata.file_information import FileContent, FileInformation, SourceFile
from ms_deisotope.data_source.scan.base import PrecursorInformation, ScanBase, ScanBunch
from ms_deisotope.data_source.scan.loader import ScanFileMetadataBase
try:
    from collections.abc import Sequence, Mapping
except ImportError:
    from collections import Sequence, Mapping
from uuid import uuid4, UUID

import numpy as np

from ms_peak_picker import PeakIndex, PeakSet, FittedPeak

try:
    from psims.mzml import writer
except ImportError:
    writer = None


from ms_deisotope import version as lib_version
from ms_deisotope.peak_set import (DeconvolutedPeak, DeconvolutedPeakSet, Envelope, IonMobilityDeconvolutedPeak)
from ms_deisotope.averagine import neutral_mass
from ms_deisotope.qc.isolation import CoIsolation

from ms_deisotope.data_source.metadata.cv import Term
from ms_deisotope.data_source.common import (
    ChargeNotProvided,
    _SingleScanIteratorImpl,
    _InterleavedGroupedScanIteratorImpl)
from ms_deisotope.data_source.metadata import data_transformation
from ms_deisotope.data_source.metadata.software import (Software)
from ms_deisotope.data_source.mzml import MzMLLoader
from ms_deisotope.data_source.scan.mobility_frame import FrameBase, Generic3DIonMobilityFrameSource, RawDataArrays3D, ProcessedIonMobilityFrame

from ms_deisotope.feature_map import ExtendedScanIndex
from ms_deisotope.feature_map.feature_fit import DeconvolutedLCMSFeature
from ms_deisotope.feature_map.feature_map import DeconvolutedLCMSFeatureMap, LCMSFeature, LCMSFeatureMap


from .common import ScanSerializerBase, ScanDeserializerBase, SampleRun, LCMSMSQueryInterfaceMixin
from .text_utils import (envelopes_to_array, decode_envelopes)


def _param(key: Union[str, Term, Any]) -> str:
    """Try to coerce a name into something we can convert into a controlled vocabulary parameter in :mod:`psims`.

    Parameters
    ----------
    key : :class:`str`, :class:`~.Term`, or other types
        The object to convert into a CV param

    Returns
    -------
    identifier : :class:`str`
        The converted identifier
    """
    if not isinstance(key, str):
        try:
            key = key.accession
        except AttributeError:
            try:
                key = key.name
            except AttributeError:
                key = str(key)
    return key


class SpectrumDescription(Sequence):
    """A helper class to calculate properties of a spectrum derived from their peak data or raw signal."""

    descriptors: List[Dict[str, float]]

    def __init__(self, attribs=None):
        Sequence.__init__(self)
        self.descriptors = list(attribs or [])

    def __getitem__(self, i):
        if isinstance(i, int):
            return self.descriptors[i]
        else:
            for d in self:
                if i == d.get('name'):
                    return d.get('value')
            raise KeyError(i)

    def __len__(self):
        return len(self.descriptors)

    def append(self, desc: dict):
        """Add the descriptor `desc` to the collection.

        Adds a descriptor which conforms to any of :mod:`psims`'s
        cvParam specification patterns.

        Parameters
        ----------
        desc : :class:`dict`, :class:`tuple`, or :class:`str`
            The descriptor to add.

        Returns
        -------
        int:
            The current size of the description list
        """
        i = len(self)
        self.descriptors.append(desc)
        return i

    def __repr__(self):
        return "{self.__class__.__name__}({self.descriptors})".format(self=self)

    @classmethod
    def from_peak_set(cls, peak_list):
        """Calculate the spectrum's descriptors from a :class:`Sequence` of :class:`~.PeakLike` objects.

        Parameters
        ----------
        peak_list : :class:`Sequence`
            The peaks to calculate properties from.

        Returns
        -------
        :class:`SpectrumDescription`
        """
        descriptors = cls()
        try:
            base_peak = max(peak_list, key=lambda x: x.intensity)
        except ValueError:
            base_peak = None
        descriptors.append({
            "name": "base peak m/z",
            "value": base_peak.mz if base_peak else 0,
        })
        descriptors.append({
            "name": "base peak intensity",
            "value": base_peak.intensity if base_peak else 0,
            "unit_name": writer.DEFAULT_INTENSITY_UNIT
        })
        descriptors.append({
            "name": "total ion current",
            "value": sum(p.intensity for p in peak_list),
            "unit_name": writer.DEFAULT_INTENSITY_UNIT
        })
        peaks_mz_order = sorted(peak_list, key=lambda x: x.mz)
        try:
            descriptors.append({
                "name": "lowest observed m/z",
                "value": peaks_mz_order[0].mz
            })
            descriptors.append({
                "name": "highest observed m/z",
                "value": peaks_mz_order[-1].mz
            })
        except IndexError:
            pass
        return descriptors

    @classmethod
    def from_arrays(cls, arrays):
        """Calculate the spectrum's descriptors from a :class:`RawDataArrays` instance.

        Parameters
        ----------
        arrays : :class:`RawDataArrays`
            The signal to calculate properties from.

        Returns
        -------
        :class:`SpectrumDescription`
        """
        descriptors = cls()
        try:
            base_peak_i = np.argmax(arrays.intensity)
        except ValueError:
            base_peak_i = None

        descriptors.append({
            "name": "base peak m/z",
            "value": arrays.mz[base_peak_i] if base_peak_i else 0
        })
        descriptors.append({
            "name": "base peak intensity",
            "value": arrays.intensity[base_peak_i] if base_peak_i else 0,
            "unit_name": writer.DEFAULT_INTENSITY_UNIT
        })
        descriptors.append({
            "name": "total ion current",
            "value": arrays.intensity.sum(),
            "unit_name": writer.DEFAULT_INTENSITY_UNIT
        })
        try:
            descriptors.append({
                "name": "lowest observed m/z",
                "value": arrays.mz[0]
            })
            descriptors.append({
                "name": "highest observed m/z",
                "value": arrays.mz[-1]
            })
        except IndexError:
            pass
        return descriptors


class MzMLSerializer(ScanSerializerBase):
    """Write :mod:`ms_deisotope` data structures to a file in mzML format.

    Attributes
    ----------
    base_peak_chromatogram_tracker : :class:`OrderedDict`
        Accumulated mapping of scan time to base peak intensity. This is
        used to write the *base peak chromatogram*.
    chromatogram_queue : :class:`list`
        Accumulate chromatogram data structures which will be written out
        after all spectra have been written to file.
    compression : :class:`str`
        The compression type to use for binary data arrays. Should be one of
        :obj:`"zlib"`, :obj:`"none"`, or :obj:`None`
    data_encoding : :class:`dict` or :class:`int` or :obj:`numpy.dtype` or :class:`str`
        The encoding specification to specify the binary encoding of numeric data arrays
        that is passed to :meth:`~.MzMLWriter.write_spectrum` and related methods.
    data_processing_list : :class:`list`
        List of packaged :class:`~.DataProcessingInformation` to write out
    deconvoluted : bool
        Indicates whether the translation should include extra deconvolution information
    file_contents_list : :class:`list`
        List of terms to include in the :obj:`<fileContents>` tag
    handle : file-like
        The file-like object being written to
    indexer : :class:`~.ExtendedScanIndex`
        The external index builder
    instrument_configuration_list : :class:`list`
        List of packaged :class:`~.InstrumentInformation` to write out
    n_spectra : int
        The number of spectra to provide a size for in the :obj:`<spectrumList>`
    processing_parameters : :class:`list`
        List of additional terms to include in a newly created :class:`~.DataProcessingInformation`
    sample_list : :class:`list`
        List of :class:`~.SampleRun` objects to write out
    sample_name : :class:`str`
        Default sample name
    sample_run : :class:`~.SampleRun`
        Description
    software_list : :class:`list`
        List of packaged :class:`~.Software` objects to write out
    source_file_list : :class:`list`
        List of packaged :class:`~.SourceFile` objects to write out
    total_ion_chromatogram_tracker : :class:`OrderedDict`
        Accumulated mapping of scan time to total intensity. This is
        used to write the *total ion chromatogram*.
    writer : :class:`~psims.mzml.writer.MzMLWriter`
        The lower level writer implementation
    """

    try:
        default_compression = writer.COMPRESSION_ZLIB
    except AttributeError:
        default_compression = None

    file_extensions = {
        "mzml",
        "mzml.gz",
    }

    default_data_encoding = {
        writer.MZ_ARRAY: np.float64,
        writer.INTENSITY_ARRAY: np.float32,
        writer.CHARGE_ARRAY: np.int32,
    }

    _format_conversion_term = "Conversion to mzML"

    handle: Union[io.IOBase, os.PathLike, str]
    n_spectra: int
    compression: Optional[str]
    data_encoding: Union[np.dtype, Dict[str, np.dtype]]
    writer: "writer.MzMLWriter"
    deconvoluted: bool
    sample_name: Optional[str]
    indexer: ExtendedScanIndex
    sample_run: SampleRun

    file_contents_list: List[FileContent]
    software_list: List[Software]
    source_file_list: List[SourceFile]
    data_processing_list: List[data_transformation.DataProcessingInformation]
    instrument_configuration_list: List[InstrumentInformation]
    sample_list: List[SampleRun]
    processing_parameters: List[data_transformation.ProcessingMethod]

    _run_tag: Optional[ContextManager]
    _spectrum_list_tag: Optional[ContextManager]
    _chromatogram_list_tag: Optional[ContextManager]

    _this_software: Optional[Software]

    _include_software_entry: bool
    _has_started_writing_spectra: bool
    _should_close: bool


    def __init__(self, handle, n_spectra=int(2e5), compression=None,
                 deconvoluted=True, sample_name=None, build_extra_index=True,
                 data_encoding=None, include_software_entry=True, close=None):
        if data_encoding is None:
            data_encoding = self.default_data_encoding
        if writer is None:
            raise ImportError(
                "Cannot write mzML without psims. Please install psims to use this feature.")
        if compression is None:
            compression = self.default_compression
        super(MzMLSerializer, self).__init__()
        self.handle = handle
        self.n_spectra = n_spectra
        self.compression = compression
        self.data_encoding = data_encoding
        self._has_started_writing_spectra = False
        self._should_close = close
        self.writer = self._make_writer(handle)
        self.writer.begin()
        self._run_tag = None
        self._spectrum_list_tag = None
        self._chromatogram_list_tag = None

        self.writer.controlled_vocabularies()
        self.deconvoluted = deconvoluted

        self._initialize_description_lists()
        self._init_sample(sample_name)

        self.total_ion_chromatogram_tracker = OrderedDict()
        self.base_peak_chromatogram_tracker = OrderedDict()
        self.chromatogram_queue = []

        self.indexer = None
        if build_extra_index:
            self.indexer = ExtendedScanIndex()
        self._include_software_entry = include_software_entry
        self._this_software = None

    def _make_writer(self, handle):
        return writer.MzMLWriter(handle, close=self._should_close)

    def _init_sample(self, sample_name, **kwargs):
        self.sample_name = sample_name
        self.sample_run = SampleRun(name=sample_name, uuid=str(uuid4()))
        self.add_sample({
            "name": self.sample_run.name,
            "id": "sample_1",
            "params": [
                {"name": "SampleRun-UUID", "value": self.sample_run.uuid},
            ]})

    def _initialize_description_lists(self):
        self.file_contents_list = []
        self.software_list = []
        self.source_file_list = []
        self.data_processing_list = []
        self.instrument_configuration_list = []
        self.sample_list = []

        self.processing_parameters = []

    def add_instrument_configuration(self, configuration: InstrumentInformation):
        """Add an :class:`~.InstrumentInformation` object to the output document.

        Parameters
        ----------
        configuration: :class:`~.InstrumentInformation`
            The instrument configuration to add
        """
        component_list = []
        for group in configuration.groups:
            tag = None
            if group.type == 'source':
                tag = self.writer.Source
            elif group.type == 'analyzer':
                tag = self.writer.Analyzer
            elif group.type == 'detector':
                tag = self.writer.Detector
            else:
                continue
            component_list.append(
                tag(order=group.order, params=[g.name for g in group]))

        params = []
        instrument_model = configuration.model
        if instrument_model:
            if isinstance(instrument_model, InstrumentModel):
                instrument_model = instrument_model.name
            params.append(instrument_model)
        instrument_serial = configuration.serial_number
        if instrument_serial:
            params.append({
                "name": "instrument serial number",
                "value": instrument_serial
            })

        conf_software = configuration.software
        if conf_software is not None:
            if isinstance(conf_software, Software):
                conf_software = conf_software.id
        config_element = self.writer.InstrumentConfiguration(
            configuration.id, component_list,
            params=params,
            software_reference=conf_software)

        self.instrument_configuration_list.append(config_element)

    def add_software(self, software_description: Software):
        """Add a :class:`~.Software` object to the output document.

        Parameters
        ----------
        software_description : :class:`~.Software`
            The software description to add
        """
        self.software_list.append(software_description)

    def add_file_information(self, file_information: FileInformation):
        """Add the information of a :class:`~.FileInformation` to the output document.

        Parameters
        ----------
        file_information: :class:`~.FileInformation`
            The information to add.
        """
        for key, value in file_information.contents.items():
            if value is None:
                value = ''
            self.add_file_contents({str(key): value})
        for source_file in file_information.source_files:
            self.add_source_file(source_file)

    def add_file_contents(self, file_contents: Union[str, Mapping, FileContent]):
        """Add a key to the resulting :obj:`<fileDescription>` of the output document.

        Parameters
        ----------
        file_contents: :class:`str` or :class:`Mapping`
            The parameter to add
        """
        self.file_contents_list.append(file_contents)

    def remove_file_contents(self, name: Union[str, Mapping, FileContent]):
        """Remove a key to the resulting :obj:`<fileDescription>` of the output document.

        Parameters
        ----------
        file_contents: :class:`str` or :class:`Mapping`
            The parameter to remove

        Raises
        ------
        KeyError:
            When the content is not found.
        """
        i = None
        for i, content in enumerate(self.file_contents_list):
            if isinstance(content, Mapping):
                if 'name' in content:
                    content = content['name']
                elif len(content) == 1:
                    content = list(content.keys())[0]
                else:
                    continue
            if content == name:
                break
        else:
            raise KeyError(name)
        if i is None:
            raise KeyError(name)
        self.file_contents_list.pop(i)

    def add_source_file(self, source_file: SourceFile):
        """Add the :class:`~.SourceFile` to the output document.

        Parameters
        ----------
        source_file : :class:`~.SourceFile`
            The source fil to add
        """
        unwrapped = {
            "name": source_file.name,
            "location": source_file.location,
            "id": source_file.id,
            "params": []
        }

        for key, value in source_file.parameters.items():
            accession = getattr(key, 'accession', str(key))
            if accession is None:
                accession = str(key)
            unwrapped['params'].append((accession, value))

        if source_file.id_format:
            unwrapped['params'].append(str(source_file.id_format))
        if source_file.file_format:
            unwrapped['params'].append(str(source_file.file_format))
        self.source_file_list.append(unwrapped)

    def add_data_processing(self, data_processing_description: Union[data_transformation.DataProcessingInformation,
                                                                     data_transformation.ProcessingMethod]):
        """Add a new :class:`~.DataProcessingInformation` or :class:`~ProcessingMethod`.

        Creates a new :obj:`<dataProcessing>` entry describing one or more :obj:`<processingMethod>`s
        for a single referenced :class:`~.Software` instance.

        Parameters
        ----------
        data_processing_description : :class:`~.DataProcessingInformation` or :class:`~.ProcessingMethod`
            Data manipulation sequence to add to the document
        """
        if isinstance(data_processing_description, data_transformation.DataProcessingInformation):
            methods = []
            for method in data_processing_description:
                content = []
                for op, val in method:
                    content.append({'name': op.name, 'value': val})
                method_descr = {
                    'software_reference': method.software_id,
                    'order': method.order,
                    'params': content
                }
                methods.append(method_descr)
            payload = {
                'id': data_processing_description.id,
                'processing_methods': methods
            }
            self.data_processing_list.append(payload)
        elif isinstance(data_processing_description, data_transformation.ProcessingMethod):
            content = []
            for op, val in data_processing_description:
                if isinstance(val, list):
                    for v in val:
                        content.append({"name": op.name, 'value': v})
                else:
                    content.append({"name": op.name, 'value': val})
            payload = {
                'id': "data_processing_%d" % len(self.data_processing_list),
                'processing_methods': [{
                    'software_reference': data_processing_description.software_id,
                    'order': data_processing_description.order,
                    'params': content
                }]
            }
            self.data_processing_list.append(payload)
        else:
            self.data_processing_list.append(data_processing_description)

    def add_processing_parameter(self, name: str, value: Optional[Union[str, int, float]]=None):
        """Add a new processing method to the writer's own :obj:`<dataProcessing>` element.

        Parameters
        ----------
        name : str
            The processing technique's name
        value : obj
            The processing technique's value, if any
        """
        self.processing_parameters.append({"name": name, "value": value})

    def add_sample(self, sample: SampleRun):
        self.sample_list.append(sample)

    def copy_metadata_from(self, reader: ScanFileMetadataBase):
        """Copies the file-level metadata from an instance of :class:`~.ScanFileMetadataBase` into the metadata.

        Parameters
        ----------
        reader : :class:`~.ScanFileMetadataBase`
            The source to copy metadata from
        """
        try:
            description = reader.file_description()
            self.add_file_information(description)
        except AttributeError:
            pass

        try:
            instrument_configs = reader.instrument_configuration()
        except AttributeError:
            instrument_configs = []
        for config in instrument_configs:
            self.add_instrument_configuration(config)

        try:
            software_list = reader.software_list()
        except AttributeError:
            software_list = []
        for software in software_list:
            self.add_software(software)

        try:
            data_processing_list = reader.data_processing()
        except AttributeError:
            data_processing_list = []
        for data_processing_ in data_processing_list:
            self.add_data_processing(data_processing_)

    def _create_file_description(self):
        self.writer.file_description(
            self.file_contents_list, self.source_file_list)

    def _create_software_list(self):
        software_list = []
        if self._include_software_entry:
            if self._this_software is None:
                self._this_software = this_software = self._make_software_entry()
            else:
                this_software = self._this_software
            self.software_list.append(this_software)
        for sw in self.software_list:
            d = {
                'id': sw.id,
                'version': sw.version
            }
            if sw.is_name(sw.name):
                d[sw.name] = ''
            else:
                d['MS:1000799'] = sw.name
            d['params'] = list(sw.options.items())
            software_list.append(d)

        self.writer.software_list(software_list)

    def _make_software_entry(self):
        ms_deisotope_entries = []
        for sw in self.software_list:
            if 'ms_deisotope' in str(sw.id):
                ms_deisotope_entries.append(str(sw.id))
        for i in range(1, 100):
            query = 'ms_deisotope_%d' % i
            if query in ms_deisotope_entries:
                continue
            else:
                new_entry_id = query
                break
        else:
            new_entry_id = 'ms_deisotope_%s' % str(uuid4())
        self._this_software = inst = Software("ms_deisotope", new_entry_id, lib_version)
        return inst

    def _create_sample_list(self):
        self.writer.sample_list(self.sample_list)

    def build_processing_method(self, order=1, picked_peaks=True, smoothing=True,
                                baseline_reduction=True, additional_parameters=tuple(),
                                software_id=None, data_processing_id=None):
        if software_id is None:
            if self._this_software is None:
                self._make_software_entry()
            software_id = self._this_software.id
        if data_processing_id is None:
            data_processing_id = 'ms_deisotope_processing_%d' % len(
                self.data_processing_list)

        method = data_transformation.ProcessingMethod(software_id=software_id)
        if self.deconvoluted:
            method.add("deisotoping")
            method.add("charge deconvolution")
            method.add("precursor recalculation")

        if picked_peaks:
            method.add("peak picking")
        if smoothing:
            method.add("smoothing")
        if baseline_reduction:
            method.add("baseline reduction")

        method.add(self._format_conversion_term)
        method.update(additional_parameters)
        method.update(self.processing_parameters)
        method.order = order
        data_processing_info = data_transformation.DataProcessingInformation(
            [method], data_processing_id)
        # self.add_data_processing(data_processing_info)
        return data_processing_info

    def _create_data_processing_list(self):
        self.writer.data_processing_list(self.data_processing_list)

    def _create_instrument_configuration(self):
        self.writer.instrument_configuration_list(
            self.instrument_configuration_list)

    def _add_spectrum_list(self):
        self._create_file_description()
        self._create_sample_list()
        self._create_software_list()
        self._create_instrument_configuration()
        self._create_data_processing_list()

        self._run_tag = self.writer.run(
            id=self.sample_name or 1,
            sample='sample_1')
        self._run_tag.__enter__()
        self._spectrum_list_tag = self.writer.spectrum_list(
            count=self.n_spectra)
        self._spectrum_list_tag.__enter__()

    def has_started_writing_spectra(self):
        return self._has_started_writing_spectra

    def _pack_activation(self, activation_information: ActivationInformation) -> dict:
        """Pack :class:`~.ActivationInformation` into a :class:`dict` structure
        which that :class:`~psims.mzml.writer.MzMLWriter` expects.

        Parameters
        ----------
        activation_information: :class:`~.ActivationInformation`

        Returns
        -------
        :class:`dict`
        """
        params = []
        params.append({
            "name": str(activation_information.method),
        })
        if activation_information.is_multiple_dissociation():
            for method in activation_information.methods[1:]:
                params.append({"name": str(method)})
        # NOTE: Only correct for CID/HCD spectra with absolute collision energies, but that is all I have
        # to test with.
        params.append({
            "name": "collision energy",
            "value": activation_information.energy,
            "unit_name": "electronvolt",
            'unit_accession': 'UO:0000266'
        })
        if activation_information.is_multiple_dissociation():
            energies = activation_information.energies[1:]
            supplemental_energy = None
            if activation_information.has_supplemental_dissociation():
                supplemental_energy = energies[-1]
                energies = energies[:-1]
            for energy in energies:
                params.append({
                    "name": "collision energy",
                    "value": energy,
                    "unit_name": "electronvolt",
                    'unit_accession': 'UO:0000266'
                })
            if supplemental_energy is not None:
                params.append({
                    "name": 'supplemental collision energy',
                    "value": supplemental_energy,
                    "unit_name": "electronvolt",
                    'unit_accession': 'UO:0000266'
                })

        for key, val in activation_information.data.items():
            arg = {
                "name": key,
                "value": val
            }
            try:
                arg['unitName'] = val.unit_info
            except AttributeError:
                pass
            params.append(arg)
        return params

    def _pack_precursor_information(self, precursor_information: PrecursorInformation,
                                    activation_information: Optional[ActivationInformation]=None,
                                    isolation_window: Optional[IsolationWindow]=None):
        """Repackage the :class:`~.PrecursorInformation`, :class:`~.ActivationInformation`,
        and :class:~.IsolationWindow` into the nested :class:`dict` structure that
        :class:`~psims.mzml.writer.MzMLWriter` expects.

        Parameters
        ----------
        precursor_information : :class:`~.PrecursorInformation`
        activation_information : :class:`~.ActivationInformation`, optional
        isolation_window : :class:`~.IsolationWindow`, optional

        Returns
        -------
        :class:`dict`
        """
        package = {}
        # If the scan bunch has been fully deconvoluted and it's PrecursorInformation
        # filled in, its extracted fields will be populated and should be used, otherwise
        # use the default read values.
        if precursor_information is not None:
            extracted_neutral_mass = precursor_information.extracted_neutral_mass
            if (extracted_neutral_mass != 0):
                package = {
                    "mz": precursor_information.extracted_mz,
                    "intensity": precursor_information.extracted_intensity,
                    "charge": precursor_information.extracted_charge,
                    "scan_id": precursor_information.precursor_scan_id,
                    "params": [
                        {"ms_deisotope:defaulted": precursor_information.defaulted},
                        {"ms_deisotope:orphan": precursor_information.orphan}
                    ]
                }
            else:
                package = {
                    "mz": precursor_information.mz,
                    "intensity": precursor_information.intensity,
                    "charge": precursor_information.charge,
                    "scan_id": precursor_information.precursor_scan_id,
                    "params": []
                }
            # This implicitly captures ion mobility which is stored as an annotation key-value pair.
            for key, value in precursor_information.annotations.items():
                key = _param(key)
                if isinstance(value, (list, tuple)):
                    for v in value:
                        package['params'].append({
                            key: v
                        })
                else:
                    package['params'].append({
                        key: value
                    })
            if precursor_information.coisolation:
                for p in precursor_information.coisolation:
                    package['params'].append({
                        "name": "ms_deisotope:coisolation",
                        "value": "%f %f %d" % (p.neutral_mass, p.intensity, p.charge)
                    })
        else:
            package['mz'] = None
            package["charge"] = None
        if package['charge'] == ChargeNotProvided:
            package["charge"] = None
        if activation_information is not None:
            package['activation'] = self._pack_activation(
                activation_information)
        if isolation_window is not None:
            package['isolation_window_args'] = {
                "lower": isolation_window.lower,
                "target": isolation_window.target,
                "upper": isolation_window.upper
            }
        return package

    def _prepare_extra_arrays(self, scan: ScanBase, **kwargs):
        deconvoluted = kwargs.get("deconvoluted", self.deconvoluted)
        extra_arrays = []
        if deconvoluted:
            score_array = [
                peak.score for peak in scan.deconvoluted_peak_set
            ]
            extra_arrays.append(("deconvolution score array", score_array))
            envelope_array = envelopes_to_array(
                [peak.envelope for peak in scan.deconvoluted_peak_set])
            extra_arrays.append(("isotopic envelopes array", envelope_array))
        else:
            arrays = scan.arrays
            if isinstance(arrays, RawDataArrays3D):
                extra_arrays.append((arrays.ion_mobility_array_type or "raw ion mobility array", arrays.ion_mobility))
            if scan.arrays.data_arrays:
                extra_arrays.extend(sorted(scan.arrays.data_arrays.items()))
        return extra_arrays

    def _get_annotations(self, scan: ScanBase) -> List[Dict[str, Any]]:
        skip = {'filter string', 'base peak intensity', 'base peak m/z', 'lowest observed m/z',
                'highest observed m/z', 'total ion current', }
        annotations = []
        for key, value in scan.annotations.items():
            if key in skip:
                continue
            annotations.append({
                key: value
            })
        return annotations

    def _get_peak_data(self, scan: ScanBase, kwargs: Mapping):
        deconvoluted = kwargs.get("deconvoluted", self.deconvoluted)
        if deconvoluted:
            centroided = True
            peak_data = scan.deconvoluted_peak_set
        elif scan.peak_set is not None:
            centroided = True
            peak_data = scan.peak_set
        else:
            centroided = not scan.is_profile
            peak_data = scan.arrays
        if deconvoluted:
            charge_array = [p.charge for p in peak_data]
        else:
            charge_array = None

        if centroided:
            descriptors = SpectrumDescription.from_peak_set(peak_data)
            mz_array = [p.mz for p in peak_data]
            intensity_array = [p.intensity for p in peak_data]
        else:
            descriptors = SpectrumDescription.from_arrays(peak_data)
            mz_array = peak_data.mz
            intensity_array = peak_data.intensity

        other_arrays = self._prepare_extra_arrays(scan, deconvoluted=deconvoluted)

        return (centroided, descriptors, mz_array, intensity_array,
                charge_array, other_arrays)

    def save_scan(self, scan: ScanBase, **kwargs):
        """Write a :class:`~.Scan` to the output document as a collection of related :obj:`<spectrum>` tags.

        .. note::

            If no spectra have been written to the output document
            yet, this method will call :meth:`_add_spectrum_list` and
            writes all of the metadata lists out. After this point,
            no new document-level metadata can be added.

        Parameters
        ----------
        scan: :class:`~.Scan`
            The scan to write.
        deconvoluted: :class:`bool`
            Whether the scan to write out should include deconvolution information
        """
        if not self._has_started_writing_spectra:
            self._add_spectrum_list()
            self._has_started_writing_spectra = True

        (centroided, descriptors, mz_array, intensity_array,
         charge_array, other_arrays) = self._get_peak_data(scan, kwargs)
        polarity = scan.polarity
        try:
            instrument_config = scan.instrument_configuration
        except AttributeError:
            instrument_config = None
        if instrument_config is None:
            instrument_config_id = None
        else:
            instrument_config_id = instrument_config.id

        scan_parameters, scan_window_list = self.extract_scan_event_parameters(scan)

        if (scan.precursor_information or scan.isolation_window or scan.activation):
            precursor_information = self._pack_precursor_information(
                scan.precursor_information,
                scan.activation,
                scan.isolation_window)
        else:
            precursor_information = None

        spectrum_params = [
            {"name": "ms level", "value": scan.ms_level},
            {"name": "MS1 spectrum"} if scan.ms_level == 1 else {"name": "MSn spectrum"},
        ] + list(descriptors)

        try:
            spectrum_params.extend(self._get_annotations(scan))
        except AttributeError:
            pass

        self.writer.write_spectrum(
            mz_array, intensity_array,
            charge_array,
            id=scan.id, params=spectrum_params,
            centroided=centroided,
            polarity=polarity,
            scan_start_time=scan.scan_time,
            compression=self.compression,
            other_arrays=other_arrays,
            instrument_configuration_id=instrument_config_id,
            precursor_information=precursor_information,
            scan_params=scan_parameters,
            scan_window_list=scan_window_list,
            encoding=self.data_encoding)

        self.total_ion_chromatogram_tracker[
            scan.scan_time] = (descriptors["total ion current"])
        self.base_peak_chromatogram_tracker[
            scan.scan_time] = (descriptors["base peak intensity"])

        if kwargs.get("_include_in_index", True) and self.indexer is not None:
            self.indexer.add_scan(scan)

    def save_scan_bunch(self, bunch: ScanBunch, **kwargs):
        """Write a :class:`~.ScanBunch` to the output document as a collection of related :obj:`<spectrum>` tags.

        .. note::

            If no spectra have been written to the output document
            yet, this method will call :meth:`_add_spectrum_list` and
            writes all of the metadata lists out. After this point,
            no new document-level metadata can be added.

        Parameters
        ----------
        bunch : :class:`~.ScanBunch`
            The scan set to write.
        """
        if bunch.precursor is not None:
            self.save_scan(bunch.precursor, _include_in_index=False, **kwargs)

        for prod in bunch.products:
            self.save_scan(prod, _include_in_index=False, **kwargs)

        if self.indexer is not None:
            self.indexer.add_scan_bunch(bunch)

    def extract_scan_event_parameters(self, scan: ScanBase):
        """Package :class:`~.ScanAcquisitionInformation` into a pair of :class:`list`s that :class:`~psims.mzml.writer.MzMLWriter` expects.

        Parameters
        ----------
        scan : :class:`~.Scan`

        Returns
        -------
        scan_parameters: :class:`list`
            Parameters qualifying the scan event (:class:`dict`)
        scan_window_list: :class:`list`
            Packed pairs of scan windows (:class:`list`)
        """
        scan_parameters = []
        scan_window_list = []
        acquisition_info = scan.acquisition_information
        filter_string = scan.annotations.get("filter_string")
        if filter_string is not None:
            scan_parameters.append({"name": "filter string", "value": filter_string})
        if acquisition_info is not None and len(acquisition_info) > 0:
            scan_event = acquisition_info[0]
            if scan_event.injection_time is not None:
                scan_parameters.append({
                    "accession": 'MS:1000927', "value": scan_event.injection_time,
                    "unit_name": getattr(scan_event.injection_time, 'unit_info', None),
                })
            traits = scan_event.traits.items()
            for name, value in traits:
                param = {"name": str(name), "value": value, 'unit_name': getattr(value, 'unit_info', None)}
                scan_parameters.append(param)
            scan_window_list = list(scan_event)
        return scan_parameters, scan_window_list

    def save_chromatogram(self, chromatogram_dict: OrderedDict[float, float], chromatogram_type: str, params=None, **kwargs):
        time_array, intensity_array = zip(*chromatogram_dict.items())
        self.writer.write_chromatogram(
            time_array, intensity_array, id=kwargs.get('id'),
            chromatogram_type=chromatogram_type, compression=self.compression,
            params=params)

    def _make_default_chromatograms(self):
        d = dict(
            chromatogram=self.total_ion_chromatogram_tracker,
            chromatogram_type='total ion current chromatogram',
            id='TIC')
        if len(self.total_ion_chromatogram_tracker) > 0:
            self.chromatogram_queue.append(d)

        d = dict(
            chromatogram=self.base_peak_chromatogram_tracker,
            chromatogram_type="basepeak chromatogram",
            id='BPC')
        if len(self.base_peak_chromatogram_tracker) > 0:
            self.chromatogram_queue.append(d)

    def write_chromatograms(self):
        self._chromatogram_list_tag = self.writer.chromatogram_list(
            count=len(self.chromatogram_queue))
        with self._chromatogram_list_tag:
            for chromatogram in self.chromatogram_queue:
                self.save_chromatogram(
                    chromatogram.pop("chromatogram"),
                    **chromatogram)

    def complete(self):
        """Finish writing to the output document.

        This closes the open list tags, empties the chromatogram accumulator,
        and closes the :obj:`<mzML>` tag, and attempts to flush the output file.
        """
        if self._spectrum_list_tag is not None:
            self._spectrum_list_tag.__exit__(None, None, None)
        if self._run_tag is not None:
            self._make_default_chromatograms()
            self.write_chromatograms()
        if self._run_tag is not None:
            self._run_tag.__exit__(None, None, None)
        self.writer.__exit__(None, None, None)
        if self.indexer is not None:
            try:
                name = self.handle.name
                try:
                    with open(ExtendedScanIndex.index_file_name(name), 'w') as ixfile:
                        self.indexer.serialize(ixfile)
                except IOError as e:
                    warnings.warn(
                        "Could not write extended index file due to error %r" % (e,))
            except AttributeError:
                warnings.warn("Could not determine name to write extended index file to")

        try:
            self.writer.outfile.flush()
        except (IOError, AttributeError, ValueError):
            pass

    def format(self):
        """A method that is no longer needed."""
        pass

    def close(self):
        self.complete()
        if hasattr(self.handle, "closed"):
            if not self.handle.closed:
                try:
                    self.handle.close()
                except AttributeError:
                    pass
        else:
            try:
                self.handle.close()
            except (AttributeError, ValueError, TypeError, OSError):
                pass


MzMLScanSerializer = MzMLSerializer


def deserialize_deconvoluted_peak_set(scan_dict, include_envelopes=True):
    if include_envelopes:
        envelopes = decode_envelopes(scan_dict["isotopic envelopes array"])
    else:
        envelopes = None
    peaks = []
    mz_array = scan_dict['m/z array']
    intensity_array = scan_dict['intensity array']
    charge_array = scan_dict['charge array']
    score_array = scan_dict['deconvolution score array']
    n = len(scan_dict['m/z array'])
    for i in range(n):
        mz = mz_array[i]
        charge = charge_array[i]
        peak = DeconvolutedPeak(
            neutral_mass(mz, charge), intensity_array[i], charge=charge, signal_to_noise=score_array[i],
            index=0, full_width_at_half_max=0, a_to_a2_ratio=0, most_abundant_mass=0,
            average_mass=0, score=score_array[i], envelope=envelopes[i] if include_envelopes else None, mz=mz
        )
        peaks.append(peak)
    peaks = DeconvolutedPeakSet(peaks)
    peaks.reindex()
    return peaks


def deserialize_deconvoluted_ion_mobility_peak_set(scan_dict):
    envelopes = decode_envelopes(scan_dict["isotopic envelopes array"])
    peaks = []
    mz_array = scan_dict['m/z array']
    intensity_array = scan_dict['intensity array']
    charge_array = scan_dict['charge array']
    score_array = scan_dict['deconvolution score array']
    drift_time_array = scan_dict['mean drift time array']
    n = len(scan_dict['m/z array'])
    for i in range(n):
        mz = mz_array[i]
        charge = charge_array[i]
        peak = IonMobilityDeconvolutedPeak(
            neutral_mass(mz, charge), intensity_array[i], charge=charge, signal_to_noise=score_array[i],
            index=0, full_width_at_half_max=0, a_to_a2_ratio=0, most_abundant_mass=0,
            average_mass=0, score=score_array[i], envelope=envelopes[i], mz=mz, drift_time=drift_time_array[i],
        )
        peaks.append(peak)
    peaks = DeconvolutedPeakSet(peaks)
    peaks.reindex()
    return peaks


def deserialize_external_deconvoluted_peaks(scan_dict, fill_envelopes=True, averagine=None):
    if averagine is None:
        from ms_deisotope import peptide
        averagine = peptide
    peaks = []
    mz_array = scan_dict['m/z array']
    intensity_array = scan_dict['intensity array']
    charge_array = scan_dict['charge array']
    for i in range(len(mz_array)):
        mz = mz_array[i]
        charge = charge_array[i]
        mass = neutral_mass(mz, charge)
        intensity = intensity_array[i]
        if fill_envelopes:
            envelope = averagine.isotopic_cluster(mz, charge, truncate_after=0.8)
            envelope = Envelope([(p.mz, p.intensity) for p in envelope.scale_raw(intensity)])
        else:
            envelope = Envelope([])
        peak = DeconvolutedPeak(
            mass, intensity, charge=charge, signal_to_noise=intensity,
            index=0, full_width_at_half_max=0, a_to_a2_ratio=0, most_abundant_mass=0,
            average_mass=0, score=intensity, envelope=envelope, mz=mz
        )
        peaks.append(peak)
    peaks = DeconvolutedPeakSet(peaks)
    peaks.reindex()
    return peaks


def deserialize_peak_set(scan_dict):
    mz_array = scan_dict['m/z array']
    intensity_array = scan_dict['intensity array']
    n = len(scan_dict['m/z array'])
    peaks = []
    for i in range(n):
        peak = FittedPeak(
            mz_array[i], intensity_array[i], 1, i, i,
            0, intensity_array[i], 0, 0)
        peaks.append(peak)
    peak_set = PeakSet(peaks)
    peak_set.reindex()
    return PeakIndex(np.array([]), np.array([]), peak_set)


class PeakSetDeserializingMixin(object):
    parse_peaks = True
    parse_envelopes = True

    def deserialize_deconvoluted_peak_set(self, scan_dict):
        try:
            if "mean drift time array" in scan_dict:
                return deserialize_deconvoluted_ion_mobility_peak_set(scan_dict)
            return deserialize_deconvoluted_peak_set(scan_dict)
        except KeyError as err:
            if "charge array" in scan_dict and "isotopic envelopes array" not in scan_dict:
                return self.deserialize_external_deconvoluted_peak_set(scan_dict)
            else:
                raise err

    def deserialize_external_deconvoluted_peak_set(self, scan_dict):
        return deserialize_external_deconvoluted_peaks(scan_dict)

    def deserialize_peak_set(self, scan_dict):
        return deserialize_peak_set(scan_dict)

    def _validate(self, scan):
        if self.parse_peaks:
            return bool(scan.deconvoluted_peak_set) or bool(scan.peak_set)
        else:
            return True

    @contextmanager
    def toggle_peak_loading(self):
        self.parse_peaks = False
        self.decode_binary = False
        yield self
        self.reset()
        self.parse_peaks = True
        self.decode_binary = True

    def _precursor_information(self, scan: ScanBase) -> PrecursorInformation:
        """Returns information about the precursor ion, if any, that this scan was derived form.

        Returns `None` if this scan has no precursor ion

        Parameters
        ----------
        scan : Mapping
            The underlying scan information storage,
            usually a `dict`

        Returns
        -------
        PrecursorInformation
        """
        precursor = super(PeakSetDeserializingMixin,
                          self)._precursor_information(scan)
        if precursor is None:
            return None
        precursor.orphan = precursor.annotations.pop(
            "ms_deisotope:orphan", None) == "true"
        precursor.defaulted = precursor.annotations.pop(
            "ms_deisotope:defaulted", None) == "true"
        coisolation_params = precursor.annotations.pop(
            "ms_deisotope:coisolation", [])
        if not isinstance(coisolation_params, list):
            coisolation_params = [coisolation_params]
        coisolation = []
        for entry in coisolation_params:
            try:
                mass, intensity, charge = entry.split(" ")
            except ValueError:
                continue
            coisolation.append(
                CoIsolation(float(mass), float(intensity), int(charge)))
        precursor.coisolation = coisolation
        return precursor

    def _make_scan(self, data):
        scan = super(PeakSetDeserializingMixin, self)._make_scan(data)
        try:
            precursor_information = scan.precursor_information
            if precursor_information:
                scan.precursor_information.default()
                selected_ion_dict = self._get_selected_ion(data)
                scan.precursor_information.orphan = selected_ion_dict.get(
                    "ms_deisotope:orphan") == "true"
                scan.precursor_information.defaulted = selected_ion_dict.get(
                    "ms_deisotope:defaulted") == "true"
                scan.annotations['precursor purity'] = data.get(
                    'precursor purity', 0)
        except KeyError:
            pass
        if self.parse_peaks:
            if "m/z array" not in data:
                warnings.warn("No m/z array found for scan %r" % (scan.id, ))
                scan.peak_set = PeakIndex(np.array([]), np.array([]), PeakSet([]))
                scan.deconvoluted_peak_set = DeconvolutedPeakSet([])
            elif "charge array" in data:
                scan.peak_set = PeakIndex(np.array([]), np.array([]), PeakSet([]))
                scan.deconvoluted_peak_set = self.deserialize_deconvoluted_peak_set(
                    data)
                if self.has_extended_index() and scan.id in self.extended_index.ms1_ids:
                    chosen_indices = self.extended_index.ms1_ids[
                        scan.id]['msms_peaks']
                    for ix in chosen_indices:
                        scan.deconvoluted_peak_set[ix].chosen_for_msms = True
            else:
                scan.peak_set = self.deserialize_peak_set(data)
                scan.deconvoluted_peak_set = None
        else:
            scan.peak_set = None
            scan.deconvoluted_peak_set = None
        packed = scan.pack(bind=True)
        return packed

    # These methods have to be replicated over from LCMSQueryInterface because of the extra
    # round of overloading `_make_scan`

    def iter_scan_headers(self, iterator=None, grouped=True):
        try:
            if not self._has_ms1_scans():
                grouped = False
        except Exception:
            pass
        self.reset()
        if iterator is None:
            iterator = iter(self._source)

        _make_scan = super(PeakSetDeserializingMixin, self)._make_scan
        _validate = super(PeakSetDeserializingMixin, self)._validate

        if grouped:
            impl = _InterleavedGroupedScanIteratorImpl(
                iterator, _make_scan, _validate)
        else:
            impl = _SingleScanIteratorImpl(iterator, _make_scan, _validate)

        for x in impl:
            yield x

        self.reset()

    def get_scan_header_by_id(self, scan_id: str) -> ScanBase:
        """Retrieve the scan object for the specified scan id. If the
        scan object is still bound and in memory somewhere, a reference
        to that same object will be returned. Otherwise, a new object will
        be created.

        Parameters
        ----------
        scan_id : str
            The unique scan id value to be retrieved

        Returns
        -------
        Scan
        """
        try:
            packed = super(PeakSetDeserializingMixin, self)._make_scan(
                self._source.get_by_id(scan_id))
            return packed
        except AttributeError as ae:
            raise AttributeError("Could not read attribute (%s) while looking up scan %s" % (
                ae, scan_id))


class ProcessedMzMLLoader(PeakSetDeserializingMixin, MzMLLoader, ScanDeserializerBase, LCMSMSQueryInterfaceMixin):
    """Extends :class:`.MzMLLoader` to support deserializing preprocessed data and to provide indexing information.

    Attributes
    ----------
    extended_index: :class:`~.ExtendedIndex`
        Holds the additional indexing information
        that may have been generated with the data
        file being accessed.
    sample_run: :class:`SampleRun`

    """

    file_extensions = {
        "mzml",
        "mzml.gz",
    }

    def __init__(self, source_file, use_index=True, use_extended_index=True):
        super(ProcessedMzMLLoader, self).__init__(source_file, use_index=use_index, decode_binary=True)
        self.extended_index = None
        self._scan_id_to_rt = dict()
        self._sample_run = None
        self._use_extended_index = use_extended_index
        if self._use_index:
            if self._use_extended_index:
                self.require_extended_index()

    def _dispose(self):
        self._scan_id_to_rt.clear()
        self.extended_index.clear()
        super(ProcessedMzMLLoader, self)._dispose()

    def __reduce__(self):
        return self.__class__, (self.source_file, self._use_index, self._use_extended_index)

    def _make_sample_run(self):
        samples = self.samples()
        sample = samples[0]
        uuid_from_sample = None
        try:
            uuid_from_sample = sample['SampleRun-UUID']
        except KeyError:
            try:
                source_name = self.source_file_name
            except AttributeError:
                source_name = str(self)
            uuid_from_sample = str(UUID(hashlib.new('md5', source_name).hexdigest()))
        return SampleRun(name=sample.name, uuid=uuid_from_sample, **dict(sample.items()))

    @property
    def sample_run(self):
        if self._sample_run is None:
            self._sample_run = self._make_sample_run()
        return self._sample_run


ProcessedMzMLDeserializer = ProcessedMzMLLoader

try:
    has_c = True
    _deserialize_deconvoluted_peak_set = deserialize_deconvoluted_peak_set
    _deserialize_peak_set = deserialize_peak_set
    from ms_deisotope._c.utils import deserialize_deconvoluted_peak_set, deserialize_peak_set
except ImportError:
    has_c = False


def extracted_features_to_3d_arrays(features: LCMSFeatureMap) -> Tuple[np.ndarray]:
    """Convert a :class:`~.LCMSFeatureMap` into a set of parallel arrays.

    The output arrays are ordered first m/z and ion mobility.

    Parameters
    ----------
    features : LCMSFeatureMap
        The ion mobility features to decompose.

    Returns
    -------
    mz_array : np.ndarray
    intensity_array : np.ndarray
    ion_mobility_array : np.ndarray
    feature_id_array : np.ndarray
    """
    mz_array = array.array('d')
    intensity_array = array.array('d')
    ion_mobility_array = array.array('d')
    feature_id_array = array.array('L')
    for i, feature in enumerate(features):
        for node in feature:
            time = node.time
            for peak in node.members:
                ion_mobility_array.append(time)
                mz_array.append(peak.mz)
                intensity_array.append(peak.intensity)
                feature_id_array.append(i)
    mz_array = np.array(mz_array, copy=False)
    intensity_array = np.array(intensity_array, copy=False)
    ion_mobility_array = np.array(ion_mobility_array, copy=False)
    feature_id_array = np.array(feature_id_array, copy=False)
    mask = np.lexsort(np.stack((ion_mobility_array, mz_array)))
    return (mz_array[mask], intensity_array[mask], ion_mobility_array[mask], feature_id_array[mask])


def deserialize_features(scan_dict: Dict[str, Union[np.ndarray, Any]],
                         ion_mobility_array_name: str='raw ion mobility array') -> LCMSFeatureMap:
    """Convert a dictionary of arrays into a :class:`~.LCMSFeatureMap.

    Parameters
    ----------
    scan_dict : Dict[str, np.ndarray]
        The arrays defining the feature map.
    ion_mobility_array_name : str
        The array to read for the ion mobility dimension.


    Returns
    -------
    LCMSFeatureMap
    """
    mz_array = scan_dict['m/z array']
    intensity_array = scan_dict['intensity array']
    drift_time_array = scan_dict[ion_mobility_array_name]
    feature_id_array = scan_dict['feature id array']
    n = len(mz_array)
    features = dict()
    for i in range(n):
        mz = mz_array[i]
        im = drift_time_array[i]
        inten = intensity_array[i]
        peak = FittedPeak(
            mz, inten, signal_to_noise=inten,
            index=0, full_width_at_half_max=0)
        try:
            feature = features[feature_id_array[i]]
        except KeyError:
            feature = features[feature_id_array[i]] = LCMSFeature([])
        feature.insert(peak, im)
    feature_map = LCMSFeatureMap(list(features.values()))
    return feature_map


def deconvoluted_features_to_3d_arrays(features: DeconvolutedLCMSFeatureMap) -> Tuple[np.ndarray]:
    """Convert a :class:`~.DeconvolutedLCMSFeatureMap` into a set of parallel arrays.

    The output arrays are ordered first m/z and ion mobility.

    Parameters
    ----------
    features : DeconvolutedLCMSFeatureMap
        The ion mobility features to decompose.

    Returns
    -------
    mz_array : np.ndarray
    intensity_array : np.ndarray
    charge_array : np.ndarray
    score_array : np.ndarray
    ion_mobility_array : np.ndarray
    envelope_array : np.ndarray
    feature_id_array : np.ndarray
    """
    mz_array = array.array('d')
    intensity_array = array.array('d')
    charge_array = array.array('i')
    score_array = array.array('d')
    ion_mobility_array = array.array('d')
    envelopes = []
    feature_id_array = array.array('L')
    point_count = 0
    for i, feature in enumerate(features):
        for node in feature:
            time = node.time
            for peak in node.members:
                ion_mobility_array.append(time)
                mz_array.append(peak.mz)
                intensity_array.append(peak.intensity)
                score_array.append(peak.score)
                charge_array.append(peak.charge)
                envelopes.append(peak.envelope)
                point_count += (len(peak.envelope) + 1) * 2
                feature_id_array.append(i)
    mz_array = np.array(mz_array, copy=False)
    intensity_array = np.array(intensity_array, copy=False)
    charge_array = np.array(charge_array, copy=False)
    score_array = np.array(score_array, copy=False)
    ion_mobility_array = np.array(ion_mobility_array, copy=False)
    feature_id_array = np.array(feature_id_array, copy=False)
    mask = np.lexsort(np.stack((ion_mobility_array, mz_array)))
    envelope_array = np.zeros(point_count, dtype=np.float32)
    k = 0
    for j in mask:
        for point in envelopes[j]:
            envelope_array[k] = point.mz
            envelope_array[k+1] = point.intensity
            k += 2
        k += 2
    return (mz_array[mask], intensity_array[mask], charge_array[mask],
            score_array[mask], ion_mobility_array[mask], envelope_array,
            feature_id_array[mask])


def deserialize_deconvoluted_features(scan_dict: Dict[str, Union[np.ndarray, Any]],
                                      ion_mobility_array_name: str='raw ion mobility array') -> DeconvolutedLCMSFeatureMap:
    """Convert a dictionary of arrays into a :class:`~.DeconvolutedLCMSFeatureMap.

    Parameters
    ----------
    scan_dict : Dict[str, np.ndarray]
        The arrays defining the feature map.
    ion_mobility_array_name : str
        The array to read for the ion mobility dimension.


    Returns
    -------
    DeconvolutedLCMSFeatureMap
    """
    envelopes = decode_envelopes(scan_dict["isotopic envelopes array"])
    mz_array = scan_dict['m/z array']
    intensity_array = scan_dict['intensity array']
    drift_time_array = scan_dict[ion_mobility_array_name]
    feature_id_array = scan_dict['feature id array']
    charge_array = scan_dict['charge array']
    score_array = scan_dict['deconvolution score array']
    n = len(mz_array)
    features = dict()
    for i in range(n):
        mz = mz_array[i]
        charge = charge_array[i]
        im = drift_time_array[i]
        peak = DeconvolutedPeak(
            neutral_mass(mz, charge), intensity_array[i], charge=charge, signal_to_noise=score_array[i],
            index=0, full_width_at_half_max=0, a_to_a2_ratio=0, most_abundant_mass=0,
            average_mass=0, score=score_array[i], envelope=envelopes[i], mz=mz
        )
        try:
            feature = features[feature_id_array[i]]
        except KeyError:
            feature = features[feature_id_array[i]
                               ] = DeconvolutedLCMSFeature([], charge)
        feature.insert(peak, im)
    feature_map = DeconvolutedLCMSFeatureMap(list(features.values()))
    return feature_map


class _IonMobility3DSerializerBase:
    def _get_peak_data(self, scan: FrameBase, kwargs):
        deconvoluted = kwargs.get("deconvoluted", self.deconvoluted)
        if deconvoluted:
            centroided = True
            peak_data = scan.deconvoluted_features
        elif scan.features:
            centroided = True
            peak_data = scan.features
        else:
            centroided = False
            peak_data = scan.arrays

        if centroided:
            descriptors = SpectrumDescription.from_peak_set(peak_data)
        else:
            descriptors = SpectrumDescription.from_arrays(peak_data)

        if deconvoluted:
            (mz_array, intensity_array, charge_array, score_array,
             ion_mobility_array, envelope_array, feature_id_array) = deconvoluted_features_to_3d_arrays(peak_data)
            other_arrays = [
                ('raw ion mobility array', ion_mobility_array),
                ('isotopic envelopes array', envelope_array),
                ('deconvolution score array', score_array),
                ('feature id array', feature_id_array),
            ]
        elif centroided:
            (mz_array, intensity_array, ion_mobility_array,
             feature_id_array) = extracted_features_to_3d_arrays(peak_data)
            charge_array = None
            other_arrays = [
                ('raw ion mobility array', ion_mobility_array),
                ('feature id array', feature_id_array),
            ]
        else:
            mz_array = peak_data.mz
            intensity_array = peak_data.intensity
            charge_array = None
            ion_mobility_array = peak_data.ion_mobility
            other_arrays = []
            if ion_mobility_array is not None:
                other_arrays.append(
                    ('raw ion mobility array', ion_mobility_array))
        return (centroided, descriptors, mz_array, intensity_array,
                charge_array, other_arrays)


class IonMobilityAware3DMzMLSerializer(_IonMobility3DSerializerBase, MzMLSerializer):
    """An mzML writer that knows how to serialize 3D ion mobility spectra.

    Instead of writing :class:`~.ScanBase` objects, this type writes :class:`~.FrameBase`
    objects. It is otherwise identical to :class:`MzMLSerializer`.
    """

    default_data_encoding = MzMLSerializer.default_data_encoding.copy()
    default_data_encoding.update({
        "feature id array": np.int32,
    })


class ProcessedGeneric3DIonMobilityFrameSource(Generic3DIonMobilityFrameSource):
    """A wrapper around a processed mzML file reader that properly reads deconvoluted IM-MS frames.

    Reads :class:`~.ProcessedIonMobilityFrame` objects from the underlying scan source.
    """

    def _make_frame(self, data) -> ProcessedIonMobilityFrame:
        frame = super()._make_frame(data)
        if 'feature id array' in frame._data:
            if 'charge array' in frame._data:
                frame.deconvoluted_features = deserialize_deconvoluted_features(frame._data)
            else:
                frame.features = deserialize_features(frame._data)
        return frame.pack()



# -*- coding: utf-8 -*-
"""Heuristic methods for estimating whether a peak list is likely
to produce a high quality interpretation.
"""
import numbers

import numpy as np


def xrea(peaks):
    """Calculate an approximation of the Xrea spectral quality metric.

    Xrea is essentially a measure of whether or not a spectrum is "flat", having no prominent
    peaks. Such spectra are more likely to be low quality.

    This implementation is based upon the implementation in SpectraST's denoising method [1].

    Parameters
    ----------
    peaks : :class:`Iterable` of :class:`PeakLike` or :class:`Iterable` of :class:`float`
        The peaks to calculate the quality measure of, or their intensities

    Returns
    -------
    :class:`float`

    References
    ----------
    [1] Shao, W., & Lam, H. (2013). Denoising Peptide Tandem Mass Spectra for Spectral Libraries: A Bayesian Approach.
        Journal of Proteome Research, 12(7), 3223–3232. https://doi.org/10.1021/pr400080b
    """
    n_peaks = len(peaks)
    if n_peaks == 0:
        return 0
    if isinstance(peaks, np.ndarray):
        intensities = np.array(peaks)
    else:
        try:
            intensities = np.array([p.intensity for p in peaks])
        except AttributeError:
            if isinstance(peaks[0], numbers.Number):
                intensities = np.array(peaks)
            else:
                raise

    intensities[::].sort()

    cumulative_intensity = np.cumsum(intensities)
    normalizer = np.cumsum(cumulative_intensity)[-1]
    cumulative_intensity = cumulative_intensity[-1]

    triangle = (cumulative_intensity * float(n_peaks + 1)) * 0.5
    xrea_ = (triangle - normalizer) / triangle
    return xrea_



cimport numpy as np


cdef double erf(double x)

cpdef gaussian_shape(np.ndarray[np.float64_t, ndim=1] xs, double center,
                     double amplitude, double sigma)

cpdef bigaussian_shape(np.ndarray[np.float64_t, ndim=1] xs, double center,
                       double amplitude, double sigma_left, double sigma_right)

cpdef skewed_gaussian_shape(np.ndarray[np.float64_t, ndim=1] xs, double center,
                            double amplitude, double sigma, double gamma)



"""Implementation based upon [1].

.. warning::

    This implementation is *experimental* and uses a lot of RAM.

References
----------
[1] Kretizberg, P. A., Bern, M., Shu, Q., Yang, F., & Serang, O. (2019). The alphabet projection of spectra.
    Journal of Proteome Research, acs.jproteome.9b00216. https://doi.org/10.1021/acs.jproteome.9b00216
"""

# -*- coding: utf-8 -*-


import math

from collections import defaultdict

import numpy as np
from scipy import signal

from ms_deisotope.utils import uid



def _enumerate_powers_of_2(max_value):
    max_value = int(max_value)
    bins = np.zeros(max_value)
    i = 0
    j = 0
    j2 = 2 ** j
    while i < max_value:
        while i < j2 and i < max_value:
            bins[i] = j2
            i += 1
        j += 1
        j2 = 2 ** j
    return bins


precalc_powers_of_2 = _enumerate_powers_of_2(1000)


def _search_for_nearest_power_of_2(value):
    i = 0
    while value > 2 ** i:
        i += 1
    return 2 ** i


def power_of_2(value):
    if value < len(precalc_powers_of_2):
        return precalc_powers_of_2[value]
    else:
        return _search_for_nearest_power_of_2(value)


def random_unit_vectors(n_planes, n_dims, random_state=None):
    """Generate `n_planes` random vectors for locality sensitive hashing
    of binned spectra with `n_dims` bins.

    Parameters
    ----------
    n_planes : int
        The number of vectors to generate
    n_dims : int
        The size of each random vector

    Returns
    -------
    list
    """
    if random_state is None:
        random_state = np.random
    unit_vectors = []
    for _i in range(n_planes):
        plane = random_state.uniform(-1, 1, n_dims)
        magnitude = np.linalg.norm(plane)
        plane /= magnitude
        unit_vectors.append(plane)
    return unit_vectors



class LSH(object):
    def __init__(self, n_planes, span=2000.0, epsilon=0.02, random_state=None):
        if random_state is None:
            random_state = np.random
        self.span = span
        self.epsilon = epsilon
        self.n_dims = power_of_2(int(span / epsilon))
        self.n_planes = n_planes
        self.random_state = random_state
        self.unit_vectors = self._create_cutting_planes()
        self.hash_bins = defaultdict(list)

    def bin_peaks(self, peaks):
        return BinnedPeakSet(peaks, self.span, self.epsilon, key=peaks)

    def _create_cutting_planes(self):
        return random_unit_vectors(self.n_planes, self.n_dims, self.random_state)

    def add(self, peaks):
        if not isinstance(peaks, BinnedPeakSet):
            peaks = self.bin_peaks(peaks)
        hv = peaks.hash_unit_vectors(self.unit_vectors)
        self.hash_bins[hv.sum()].append(peaks)
        return hv


class BinnedPeakSet(object):
    def __init__(self, peak_set, span=None, epsilon=0.02, key=None):
        if span is None:
            span = 0
        if key is None:
            key = uid()
        self.key = key
        self.peak_set = peak_set
        self.span = max(span, self._guess_span())
        self.epsilon = epsilon

        self.indices = None
        self.binned_signal = None
        self.discretize()

    def _guess_span(self):
        return self.peak_set[-1].mz - self.peak_set[0].mz

    def __eq__(self, other):
        if other is None:
            return False
        try:
            return self.key == other.key
        except AttributeError:
            return self.peak_set == list(other)

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return hash(self.key)

    def bin_steps(self):
        min_mz = self.peak_set[0].mz
        max_mz = self.peak_set[-1].mz
        bins_per_mz = len(self.binned_signal) / (max_mz - min_mz)
        step = int(math.ceil((max_mz - min_mz) / 100.0) * 10)
        locs = []
        labels = []
        for i in range(10):
            locs.append(bins_per_mz * i * step)
            labels.append(min_mz + i * step)
        return locs, labels

    def discretize(self, span=None):
        if span is None:
            span = self.span
        else:
            span = self.span = max(span, self._guess_span())
        num_bins = power_of_2(int(math.ceil(span / self.epsilon)))
        bins = np.linspace(0, span + self.epsilon, num=num_bins)
        min_mz = self.peak_set[0].mz
        # This ignores the possibility that a peak delta falls outside of span
        self.indices = np.digitize([peak.mz - min_mz for peak in self.peak_set], bins)
        self.binned_signal = np.zeros(num_bins)
        for i, peak in enumerate(self.peak_set):
            # The original implementation just set the bin to exp(intensity) of the last
            # peak to fall in it, but exp() overflows with numbers higher than ~700.
            #
            # The absolute magnitude of the value here doesn't matter, just that it is
            # non-zero?
            self.binned_signal[self.indices[i]] += peak.intensity

    def similarity(self, other):
        self_clamped = self.binned_signal.clip(0, 1)
        other_clamped = other.binned_signal.clip(0, 1)
        convolved = signal.fftconvolve(self_clamped, other_clamped[::-1])
        i = np.argmax(convolved)
        score = convolved[i]
        return np.roll(other.binned_signal, i + 1), i + 1, score

    def hash_unit_vectors(self, unit_vectors):
        power_spectrum = np.abs(np.fft.fft(self.binned_signal))
        power_spectrum[0] = 0.0
        value = np.zeros_like(power_spectrum)
        for i, v in enumerate(unit_vectors):
            bit = v.dot(power_spectrum)
            value[i] = (bit > 0) * 2 ** i
        return value


def discretize(peak_set, span=None, epsilon=0.02):
    return BinnedPeakSet(peak_set, span, epsilon)



# -*- coding: utf-8 -*-
"""mzMLb is a standard rich HDF5-based format for raw mass spectrometry data storage.
This module provides :class:`MzMLbLoader`, a :class:`~.RandomAccessScanSource`
implementation. It is based upon the mzML XML file format, re-using a subset of the
features. The original design for mzMLb is described in [Bhamber]_.

The parser is based on :mod:`pyteomics.mzmlb`. It requires :mod:`h5py` to be installed
for reading, and :mod:`hdf5plugin` to use the faster, non-zlib-based compressors.

References
==========
.. [Bhamber] Bhamber, R. S., Jankevics, A., Deutsch, E. W., Jones, A. R., & Dowsey, A. W. (2021).
    MzMLb: A Future-Proof Raw Mass Spectrometry Data Format Based on Standards-Compliant
    mzML and Optimized for Speed and Storage Requirements. Journal of Proteome Research,
    20(1), 172–183. https://doi.org/10.1021/acs.jproteome.0c00192
"""

import logging
logging.getLogger("hdf5plugin").addHandler(logging.NullHandler())

try:
    from pyteomics import mzmlb
    _BaseParser = mzmlb.MzMLb

except ImportError as impl_import_err:
    mzmlb = None

    class _BaseParser:
        def __init__(self, *args, **kwargs):
            raise impl_import_err

from .mzml import MzMLLoader as _MzMLLoader
from ._compression import DefinitelyFastRandomAccess


class _MzMLbParser(_BaseParser):

    def _handle_param(self, element, **kwargs):
        try:
            element.attrib["value"]
        except KeyError:
            element.attrib["value"] = ""
        return super(_MzMLbParser, self)._handle_param(element, **kwargs)


class MzMLbLoader(_MzMLLoader):
    """Reads scans from PSI-HUPO mzMLb HDF5 files. Provides both iterative and
    random access.

    Attributes
    ----------
    source_file: str
        Path to file to read from.
    source: pyteomics.mzmlb.MzMLb
        Underlying scan data source
    """
    _parser_cls = _MzMLbParser

    @property
    def has_fast_random_access(self):
        return DefinitelyFastRandomAccess

    @classmethod
    def prebuild_byte_offset_file(cls, path):
        """A stub method. MzMLb does not require an external index.

        Parameters
        ----------
        path : :class:`str` or file-like
            The path to the file to index, or a file-like object with a name attribute.
        """
        return None


def is_mzmlb_file(path):
    """Detect whether or not the file referenced by ``path``
    is a mzMLb file.

    Parameters
    ----------
    path: :class:`str`
        The path to test

    Returns
    -------
    :class:`bool`:
        Whether or not the file is a mzMLb file.
    """
    try:
        import h5py
        if mzmlb is None:
            raise impl_import_err
    except ImportError:
        return False
    try:
        source = h5py.File(path, 'r')
        source['mzML']
        return True
    except KeyError:
        return False


def infer_reader(path):
    """If the file referenced by ``path`` is a mzMLb
    file, return the callable (:class:`MzMLbLoader`) to
    open it, otherwise raise an exception.

    Parameters
    ----------
    path: :class:`str`
        The path to test

    Returns
    -------
    :class:`type`:
        The type to use to open the file

    Raises
    ------
    :class:`ValueError`:
        If the file is not a mzMLb file
    """
    if is_mzmlb_file(path):
        return MzMLbLoader
    raise ValueError("Not mzMLb File")


def determine_if_available():
    """Checks whether or not the mzMLb HDF5-based
    file reading feature is available.

    Returns
    -------
    :class:`bool`:
        Whether or not the feature is enabled.
    """
    try:
        import h5py
        if mzmlb is None:
            raise ImportError('pyteomics.mzmlb')
        return True
    except (OSError, ImportError):
        return False



import array
from typing import Any, Dict, List, Optional
import warnings
import logging

from abc import abstractmethod
from weakref import WeakValueDictionary
from itertools import chain

from collections import namedtuple, defaultdict

import numpy as np

from ms_peak_picker import average_signal, FittedPeak
from ms_deisotope.data_source.metadata.activation import ActivationInformation
from ms_deisotope.data_source.metadata.scan_traits import IsolationWindow, ScanAcquisitionInformation

try:
    from ms_peak_picker.scan_averaging import GridAverager
except ImportError:
    GridAverager = None

from ms_deisotope.data_source.scan.scan import Scan, WrappedScan

from ms_deisotope.peak_set import IonMobilityDeconvolutedPeak, DeconvolutedPeakSet
from ms_deisotope.peak_dependency_network import IntervalTreeNode, Interval

from .base import PrecursorInformation, RawDataArrays, ScanBase, ScanBunch

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

class IonMobilitySource(object):
    @abstractmethod
    def _frame_id(self, data) -> str:
        raise NotImplementedError()

    @abstractmethod
    def _frame_index(self, data) -> int:
        raise NotImplementedError()

    def _frame_time(self, data) -> float:
        scan = self.get_scan_by_index(
            self._frame_start_scan_index(data))
        return scan.scan_time

    def _frame_ms_level(self, data) -> int:
        scan = self.get_scan_by_index(
            self._frame_start_scan_index(data))
        return scan.ms_level

    @abstractmethod
    def _frame_start_scan_index(self, data) -> int:
        raise NotImplementedError()

    @abstractmethod
    def _frame_end_scan_index(self, data) -> int:
        raise NotImplementedError()

    def _frame_precursor_information(self, data):
        scan = self.get_scan_by_index(
            self._frame_start_scan_index(data))
        return scan.precursor_information

    def _frame_activation(self, data):
        scan = self.get_scan_by_index(
            self._frame_start_scan_index(data))
        return scan.activation

    def _frame_isolation_window(self, data):
        scan = self.get_scan_by_index(
            self._frame_start_scan_index(data))
        return scan.isolation_window

    def _frame_polarity(self, data) -> int:
        scan = self.get_scan_by_index(
            self._frame_start_scan_index(data))
        return scan.polarity

    def _frame_drift_times(self, data) -> List[float]:
        scans = []
        for i in range(self._frame_start_scan_index(data), self._frame_end_scan_index(data)):
            scan = self.get_scan_by_index(i)
            scans.append(scan.drift_time)
        return scans

    def _frame_scans(self, data) -> List[ScanBase]:
        scans = []
        for i in range(self._frame_start_scan_index(data), self._frame_end_scan_index(data)):
            scan = self.get_scan_by_index(i)
            scans.append(scan)
        return scans

    def _frame_annotations(self, data) -> dict:
        return {}

    def _frame_arrays(self, data) -> 'RawDataArrays3D':
        scans = self._frame_scans(data)
        return RawDataArrays3D.stack(scans)

    def _frame_ion_mobilities(self, data) -> np.ndarray:
        scans = self._frame_scans(data)
        return np.array([s.drift_time for s in scans])

    def _frame_acquisition_information(self, data):
        scan = self.get_scan_by_index(self._frame_start_scan_index(data))
        acq = scan.acquisition_information
        if acq is None:
            return acq
        acq = acq.copy()
        for event in acq:
            im = event._ion_mobility
            mt = im.ion_mobility_type()
            if mt is not None:
                im.remove_ion_mobility_type(mt)
        return acq


class IonMobilitySourceRandomAccessFrameSource(IonMobilitySource):
    @abstractmethod
    def get_frame_by_index(self, index: int) -> 'IonMobilityFrame':
        raise NotImplementedError()

    @abstractmethod
    def get_frame_by_time(self, time: float) -> 'IonMobilityFrame':
        raise NotImplementedError()

    @abstractmethod
    def get_frame_by_id(self, id: str) -> 'IonMobilityFrame':
        raise NotImplementedError()

    def _validate_frame(self, data):
        return True

    def _make_frame(self, data) -> 'IonMobilityFrame':
        return IonMobilityFrame(data, self)

    def _cache_frame(self, frame):
        pass

    @abstractmethod
    def _default_frame_iterator(self, start_index=None):
        raise NotImplementedError()

    @abstractmethod
    def make_frame_iterator(self, iterator=None, grouped=False):
        raise NotImplementedError()

    @abstractmethod
    def start_from_frame(self, scan_id=None, rt=None, index=None, require_ms1=True, grouped=True):
        raise NotImplementedError()

    def initialize_frame_cache(self):
        """Initialize a cache which keeps track of which :class:`~.IonMobilityFrame`
        objects are still in memory using a :class:`weakref.WeakValueDictionary`.

        When a frame is requested, if the frame object is found in the cache, the
        existing object is returned rather than re-read from disk.
        """
        self._frame_cache = WeakValueDictionary()

    @property
    def frame_cache(self):
        """A :class:`weakref.WeakValueDictionary` mapping used to retrieve
        frames from memory if available before re-reading them from disk.
        """
        return self._frame_cache

    @frame_cache.setter
    def frame_cache(self, value):
        self._frame_cache = value


class _upto_index(object):
    def __init__(self, values, last_value):
        self.values = values
        self.last_value = last_value
        self.index = np.searchsorted(self.values, last_value)

    def values_up_to(self, value):
        i = np.searchsorted(self.values, value)
        intermediate = self.values[self.index:i]
        if len(intermediate):
            # The slice will contain the last stored value, which we don't
            # want to include, but unless we know that we aren't skipping
            # several values on the first request, we can't exclude it automatically.
            if np.isclose(intermediate[0], self.last_value):
                intermediate = intermediate[1:]
        self.index = i
        self.last_value = value
        return intermediate

    def __call__(self, value):
        return self.values_up_to(value)


class RawDataArrays3D(namedtuple("RawDataArrays3D", ['mz', 'intensity', 'ion_mobility'])):
    """Represent the m/z, intensity, and ion mobility arrays associated with a raw
    ion mobility frame of mass spectra.

    Thin wrapper around a ``namedtuple``, so this object supports
    the same interfaces as a tuple.

    Attributes
    ----------
    mz : :class:`np.ndarray`
        The m/z axis of a mass spectrum
    intensity : :class:`np.ndarray`
        The intensity measured at the corresponding m/z of a mass spectrum
    ion_mobility : :class:`np.ndarray`
        The ion mobility of each data point
    distinct_ion_mobility : :class:`np.ndarray`
        The distinct ion mobility values in the multidimensional space, independent
        of whether there are data points at that point
    ion_mobility_array_type : :class:`~.Term`
        The type of ion mobility array
    data_arrays : dict
        Any other data arrays
    """
    def __new__(cls, mz, intensity, ion_mobility, distinct_ion_mobility, ion_mobility_array_type=None, data_arrays=None):
        inst = super(RawDataArrays3D, cls).__new__(
            cls, mz, intensity, ion_mobility)
        inst.distinct_ion_mobility = distinct_ion_mobility
        inst.ion_mobility_array_type = ion_mobility_array_type
        inst.data_arrays = dict()
        if data_arrays:
            inst.data_arrays.update(data_arrays)
        return inst

    @property
    def drift_time(self):
        return self.ion_mobility

    def has_array(self, array_type):
        """Check if this array set contains an array of the
        requested type.

        This method uses the semantic lookup mechanism to test
        "is-a" relationships so if a more abstract term is used,
        a wider range of terms may be matched.

        Parameters
        ----------
        array_type : str or :class:`~.Term`
            The array type name to test.

        Returns
        -------
        bool
        """
        from ms_deisotope.data_source.metadata.scan_traits import binary_data_arrays
        try:
            term = binary_data_arrays[array_type]
        except KeyError:
            warnings.warn(
                "Array type %r could not be resolved, treating as a plain string" % (array_type, ))
            return array_type in self.binary_data_arrays
        if self.mz is not None and len(self.mz):
            k = binary_data_arrays['m/z array']
            if term.is_a(k):
                return k
        if self.intensity is not None and len(self.intensity):
            k = binary_data_arrays['intensity array']
            if term.is_a(k):
                return k
        if self.ion_mobility_array_type is not None:
            k = binary_data_arrays[self.ion_mobility_array_type]
            if term.is_a(k):
                return k
        for k in self.data_arrays:
            try:
                k = binary_data_arrays[k]
                if term.is_a(k):
                    return k
            except KeyError:
                if term == k:
                    return k
        return False

    def __copy__(self):
        inst = self.__class__(self.mz.copy(), self.intensity.copy(), self.ion_mobility.copy(),
                              self.distinct_ion_mobility.copy(), self.ion_mobility_array_type, {
            k: v.copy() for k, v in self.data_arrays.items()
        })
        return inst

    def copy(self):
        """Make a deep copy of this object.

        Returns
        -------
        :class:`RawDataArrays3D`
        """
        return self.__copy__()

    def __eq__(self, other):
        try:
            return np.allclose(
                self[0], other[0]) and np.allclose(
                    self[1], other[1]) and np.allclose(self[2], other[2])
        except ValueError:
            return False

    def __ne__(self, other):
        return not (self == other)

    def __mul__(self, i):
        return self.__class__(self.mz, self.intensity * i, self.ion_mobility,
                              self.distinct_ion_mobility, self.ion_mobility_array_type, self.data_arrays)

    def __div__(self, d):
        return self.__class__(self.mz, self.intensity / d, self.ion_mobility,
                              self.distinct_ion_mobility, self.ion_mobility_array_type, self.data_arrays)
    def __getitem__(self, i):
        if isinstance(i, int):
            return super(RawDataArrays3D, self).__getitem__(i)
        elif isinstance(i, (slice, list, tuple, np.ndarray)):
            return self._slice(i)
        else:
            return self.data_arrays[i]

    @classmethod
    def stack(cls, scans, ion_mobility_array_type=None):
        """Combine multiple :class:`~.Scan` objects or (ion mobility, :class:`~.RawDataArrays`)
        pairs into a single :class:`~.RawDataArrays3D`

        Parameters
        ----------
        scans : list of :class:`~.Scan`
            The single ion mobility point arrays to combine
        ion_mobility_array_type : :class:`~.Term` or :class:`str`
            The type of ion mobility array
        """
        mz = array.array('d')
        intensity = array.array('d')
        ion_mobility = array.array('d')
        data_arrays = defaultdict(lambda: array.array('d'))

        distinct_ion_mobility = array.array('d')
        for scan in scans:
            try:
                arrays = scan.arrays
                im = scan.drift_time
            except AttributeError:
                im = scan[0]
                arrays = scan[1]
            if im is not None:
                distinct_ion_mobility.append(im)
            if arrays.mz.size > 0:
                mz.extend(arrays.mz)
                intensity.extend(arrays.intensity)
                if im is not None:
                    ion_mobility.extend([im] * arrays.mz.size)
                for key, val in arrays.data_arrays.items():
                    data_arrays[key].extend(val)
        mz = np.array(mz)
        if len(ion_mobility) > 0:
            ion_mobility = np.array(ion_mobility)
            mask = np.lexsort(np.stack((ion_mobility, mz,)))
            return cls(mz[mask], np.array(intensity)[mask], ion_mobility[mask],
                    np.sort(distinct_ion_mobility), ion_mobility_array_type,
                    {k: np.array(v)[mask] for k, v in data_arrays.items()})
        else:
            return cls(mz, np.array(intensity), None, np.array([]), None,
                       {k: np.array(v) for k, v in data_arrays.items()})

    @classmethod
    def from_arrays(cls, mz_array, intensity_array, ion_mobility_array, ion_mobility_array_type=None, data_arrays=None):
        """Build a new :class:`~.RawDataArrays3D` from parallel arrays.

        This will sort all arrays w.r.t. m/z and ion mobility.

        Parameters
        ----------
        mz_array : np.ndarray
            The m/z array
        intensity_array : np.ndarray
            The intensity array
        ion_mobility_array : np.ndarray
            The ion mobility array
        data_arrays : dict
            Any extra data arrays

        Returns
        -------
        :class:`~.RawDataArrays3D`
        """
        mz_array = np.asarray(mz_array)
        ion_mobility_array = np.asarray(ion_mobility_array)
        distinct_ion_mobility = np.sort(np.unique(ion_mobility_array))
        mask = np.lexsort(np.stack((ion_mobility_array, mz_array)))
        if data_arrays:
            data_arrays = {
                k: np.asarray(v)[mask] for k, v in data_arrays.items()
            }
        return cls(
            mz_array[mask], np.asarray(intensity_array)[mask], ion_mobility_array[mask],
            distinct_ion_mobility, ion_mobility_array_type, data_arrays=data_arrays)


    @classmethod
    def empty(cls):
        return cls(np.array([]), np.array([]), np.array([]), np.array([]))

    def _slice(self, i, include_ion_mobility=True):
        mz = self.mz[i]
        intensity = self.intensity[i]
        data_arrays = {k: v[i] for k, v in self.data_arrays.items()}
        if not include_ion_mobility:
            return RawDataArrays(mz, intensity, data_arrays)
        return RawDataArrays3D(
            mz, intensity, self.ion_mobility[i], self.distinct_ion_mobility,
            self.ion_mobility_array_type, data_arrays)

    def unstack(self, include_empty=True):
        """Convert this 3D array into a list of (ion mobility, :class:`~.RawDataArrays`) pairs

        Parameters
        ----------
        include_empty : bool
            Whether to include ion mobility values with no signal. Defaults to :const:`True`.

        Returns
        -------
        arrays : list[tuple[float, :class:`~.RawDataArrays`]]
            The split arrays arranged by drift time
        """
        # Sort by m/z first, then by ion mobility
        mask = np.lexsort(np.stack((self.mz, self.ion_mobility)))
        sorted_dim = self.ion_mobility[mask]
        acc = []
        # re-sort all arrays along the mask, and then slice out in chunks.
        boundaries = np.where(np.diff(sorted_dim) != 0)[0]
        if len(boundaries) == 0:
            return [RawDataArrays(self.mz, self.intensity, self.data_arrays)]
        if boundaries[-1] < len(sorted_dim) - 1:
            boundaries = np.append(boundaries, [len(sorted_dim) - 1])
        last_boundary = 0
        index = _upto_index(self.distinct_ion_mobility, -1)
        for boundary in boundaries:
            drift_time = self.ion_mobility[mask[last_boundary]]
            for i in index(drift_time):
                acc.append((i, self._slice(slice(0, 0), False)))
            chunk = self._slice(mask[last_boundary:boundary + 1], False)
            acc.append((drift_time, chunk))
            last_boundary = boundary + 1
        return acc

    def grid(self):
        """Convert this 3D array into an intensity grid with
        m/z along the rows and ion mobility along the columns.

        Returns
        -------
        mz_axis : np.ndarray
            The m/z value for each row
        im_axis : np.ndarray
            The ion mobility value for each column
        intensity : np.ndarray
            A 2D array of intensity values at each axis position
        """
        mz_axis = np.unique(self.mz)
        im_axis = (self.distinct_ion_mobility)
        intensity = np.zeros((mz_axis.shape[0], im_axis.shape[0]))
        i1 = np.searchsorted(mz_axis, self.mz)
        i2 = np.searchsorted(im_axis, self.ion_mobility)
        intensity[i1, i2] += self.intensity
        return mz_axis, im_axis, intensity

    def find_mz(self, mz):
        """Find the nearest index to the query ``mz``

        Parameters
        ----------
        mz : float
            The m/z value to search for

        Returns
        -------
        int
            The index nearest to the query m/z
        """
        n = len(self.mz)
        lo = 0
        hi = n

        while hi != lo:
            mid = int((hi + lo) // 2)
            y = self.mz[mid]
            err = y - mz
            if abs(err) < 0.1:
                best_index = mid
                best_err = abs(err)
                i = mid
                while i >= 0:
                    y = self.mz[i]
                    err = y - mz
                    if err <= -0.1:
                        break
                    abs_err = abs(err)
                    if abs_err < best_err:
                        best_err = abs_err
                        best_index = i
                    i -= 1
                i = mid
                while i < n:
                    y = self.mz[i]
                    err = y - mz
                    if err >= 0.1:
                        break
                    abs_err = abs(err)
                    if abs_err < best_err:
                        best_err = abs_err
                        best_index = i
                    i += 1
                return best_index
            elif hi - lo == 1:
                return mid
            elif err > 0:
                hi = mid
            else:
                lo = mid
        return 0

    def between_mz(self, low, high):
        """Returns a slice of the arrays between ``low`` and ``high``
        m/z

        Parameters
        ----------
        low : float
            The lower bound m/z
        high : float
            The upper bound m/z

        Returns
        -------
        :class:`.RawDataArrays`
        """
        i = self.find_mz(low)
        mz_i = self.mz[i]
        while self.mz[i] == mz_i:
            i -= 1
        i += 1
        j = self.find_mz(high)
        mz_j = self.mz[j]
        while self.mz[j] == mz_j:
            j += 1
        if not (low <= self.mz[i] <= high):
            i += 1
        return self._slice(slice(i, j), True)


class FrameBase(object):

    time: float

    @property
    def scan_time(self) -> float:
        return self.time

    @property
    def ion_mobilities(self) -> np.ndarray:
        return self.source._frame_ion_mobilities(self._data)

    def scans(self) -> List[Scan]:
        return self.source._frame_scans(self._data)

    def bind(self, source: IonMobilitySource):
        self.source = source
        if self.precursor_information is not None:
            self.precursor_information.bind(source)
        return self

    def unbind(self):
        self.source = None
        if self.precursor_information:
            self.precursor_information.unbind()
        return self

    def __repr__(self):
        template = ("{self.__class__.__name__}({self.id}, index={self.index},"
                    " time={self.time}, ms_level={self.ms_level})")
        return template.format(self=self)

    def __eq__(self, other):
        if other is None:
            return False
        if not isinstance(other, FrameBase):
            return False
        if self.id != other.id:
            return False
        elif self.ms_level != other.ms_level:
            return False
        elif self.time != other.time:
            return False
        elif self.deconvoluted_features != other.deconvoluted_features:
            return False
        elif self.features != other.features:
            return False
        return True

    def __ne__(self, other):
        return not self == other

    __hash__ = None

    def to_deconvoluted_peak_set(self, time_bound=None, mass_upper_bound: float = float('inf')) -> DeconvolutedPeakSet:
        if self.deconvoluted_features is None:
            raise ValueError(
                "Must first deconvolute IMS features before converting to a time-binned peak set!")
        if time_bound is not None:
            cft = IntervalTreeNode.build([
                Interval(f.start_time, f.end_time, [f])
                for f in self.deconvoluted_features
                if f.neutral_mass <= mass_upper_bound
            ])
            features = chain.from_iterable(cft.overlaps(*time_bound))
        else:
            features = self.deconvoluted_features
        return features_to_peak_set(features)


class IonMobilityFrame(FrameBase):
    """A :class:`IonMobilityFrame` represents an single time point acquisition of
    multiple mass spectra across multiple ion mobility drift time points. Because
    of its drift time by m/z structure, it does not have 1-D peak sets, but pseudo-
    :class:`~.LCMSFeatureMap`-like data structures which conserve the over-time
    property.
    """

    source: IonMobilitySource

    def __init__(self, data, source):
        self._data = data
        self.source = source
        self.features = None
        self.deconvoluted_features = None

        self._unload()

    def _unload(self):
        self._id = None
        self._ms_level = None
        self._time = None
        self._polarity = None
        self._index = None
        self._start_scan_index = None
        self._end_scan_index = None

        self._arrays = None
        self._annotations = None

        self._isolation_window = None
        self._precursor_information = None
        self._activation = None
        self._acquisition_information = None

    @property
    def id(self) -> str:
        if self._id is None:
            self._id = self.source._frame_id(self._data)
        return self._id

    @id.setter
    def id(self, value):
        self._id = value

    @property
    def index(self) -> int:
        if self._index is None:
            self._index = self.source._frame_index(self._data)
        return self._index

    @index.setter
    def index(self, value):
        self._index = value

    @property
    def time(self) -> float:
        if self._time is None:
            self._time = self.source._frame_time(self._data)
        return self._time

    @time.setter
    def time(self, value):
        self._time = value

    @property
    def ms_level(self) -> int:
        if self._ms_level is None:
            self._ms_level = self.source._frame_ms_level(self._data)
        return self._ms_level

    @ms_level.setter
    def ms_level(self, value):
        self._ms_level = value

    @property
    def polarity(self) -> int:
        if self._polarity is None:
            self._polarity = self.source._frame_polarity(self._data)
        return self._polarity

    @polarity.setter
    def polarity(self, value):
        self._polarity = value

    @property
    def start_scan_index(self) -> int:
        if self._start_scan_index is None:
            self._start_scan_index = self.source._frame_start_scan_index(self._data)
        return self._start_scan_index

    @start_scan_index.setter
    def start_scan_index(self, value):
        self._start_scan_index = value

    @property
    def end_scan_index(self) -> int:
        if self._end_scan_index is None:
            self._end_scan_index = self.source._frame_end_scan_index(self._data)
        return self._end_scan_index

    @end_scan_index.setter
    def end_scan_index(self, value):
        self._end_scan_index = value

    @property
    def precursor_information(self) -> Optional[PrecursorInformation]:
        if self._precursor_information is None:
            self._precursor_information = self.source._frame_precursor_information(self._data)
        return self._precursor_information

    @precursor_information.setter
    def precursor_information(self, value):
        self._precursor_information = value

    @property
    def activation(self) -> Optional[ActivationInformation]:
        if self._activation is None:
            self._activation = self.source._frame_activation(self._data)
        return self._activation

    @activation.setter
    def activation(self, value):
        self._activation = value

    @property
    def isolation_window(self) -> Optional[IsolationWindow]:
        if self._isolation_window is None:
            self._isolation_window = self.source._frame_isolation_window(self._data)
        return self._isolation_window

    @isolation_window.setter
    def isolation_window(self, value):
        self._isolation_window = value

    @property
    def acquisition_information(self) -> ScanAcquisitionInformation:
        if self._acquisition_information is None:
            self._acquisition_information = self.source._frame_acquisition_information(self._data)
        return self._acquisition_information

    @acquisition_information.setter
    def acquisition_information(self, value):
        self._acquisition_information = value

    @property
    def arrays(self) -> RawDataArrays3D:
        if self._arrays is None:
            self._arrays = self.source._frame_arrays(self._data)
        return self._arrays

    @arrays.setter
    def arrays(self, value: Optional[RawDataArrays3D]):
        if value is not None and not isinstance(value, RawDataArrays3D):
            raise TypeError(f"Cannot assign type {type(value)} to arrays!")
        self._arrays = value

    @property
    def annotations(self) -> Dict[str, Any]:
        if self._annotations is None:
            self._annotations = self.source._frame_annotations(self._data)
        return self._annotations

    @annotations.setter
    def annotations(self, value):
        self._annotations = dict(value)

    def flatten_to_scan(self) -> WrappedScan:
        scans = self.scans()
        mz, intensity = average_signal([scan.arrays for scan in scans], dx=0.002)
        intensity *= len(scans)
        arrays = RawDataArrays(mz, intensity)
        return WrappedScan(scans[0]._data, self.source, arrays, id="merged=%d" % self.index)

    def get_scan_by_drift_time(self, drift_time: float) -> ScanBase:
        dt_axis = self.ion_mobilities
        lo = 0
        hi = n = len(dt_axis)
        best_match = None
        best_error = float('inf')

        # Handles infinities
        if drift_time > dt_axis[-1]:
            return self.source.get_scan_by_index(self.end_scan_index - 1)

        while hi != lo:
            mid = (hi + lo) // 2
            dt = dt_axis[mid]
            err = abs(dt - drift_time)
            if err < best_error:
                best_error = err
                best_match = mid
            if (dt == drift_time) or (hi - 1 == lo):
                i = mid - 1
                while i >= 0:
                    err = abs(dt_axis[i] - drift_time)
                    if err < best_error:
                        best_error = err
                        best_match = i
                        i -= 1
                    else:
                        break
                i = mid + 1
                while i < n:
                    err = abs(dt_axis[i] - drift_time)
                    if err < best_error:
                        best_error = err
                        best_match = i
                        i += 1
                    else:
                        break
                return self.source.get_scan_by_index(self.start_scan_index + best_match)
            elif dt > drift_time:
                hi = mid
            else:
                lo = mid

    def extract_features(self, error_tolerance=1.5e-5, max_gap_size=0.25, min_size=2, average_within=0, average_across=0, dx=0.002, num_threads=3,
                         low_memory=False, minimum_intensity=0.0, denoise=0, **kwargs) -> 'IonMobilityFrame':
        from ms_deisotope.feature_map import feature_map
        scans: List[Scan] = self.scans()
        n = len(scans)
        lff = feature_map.LCMSFeatureForest(error_tolerance=error_tolerance)
        scan: Scan
        averager = None
        if not low_memory:
            min_mz = float('inf')
            max_mz = 0
            arrays = []
            for scan in scans:
                a = scan.arrays
                arrays.append((a.mz.astype(float), a.intensity.astype(float)))
                if len(a.mz):
                    min_mz = min(a.mz[0], min_mz)
                    max_mz = max(a.mz[-1], max_mz)
            if min_mz > max_mz:
                min_mz = 0
                max_mz = 1

            averager = GridAverager(min_mz, max_mz, len(scans), dx=dx)
            averager.add_spectra(arrays, num_threads)

        for i, scan in enumerate(scans):
            if average_within:
                if not low_memory:
                    lo = max(i - average_within, 0)
                    hi = min(i + average_within + 1, len(scans))
                    array_data = RawDataArrays(
                        averager.mz_axis, averager.average_indices(lo, hi, n_workers=1))
                    acq_info = scan.acquisition_information
                    scan = WrappedScan(scan._data, scan.source, array_data, _acquisition_information=acq_info)
                else:
                    acc: List[ScanBase] = []
                    for j in range(max((i - average_within, 0)), i):
                        other_scan = scans[j]
                        acc.append(other_scan)
                    for j in range(i + 1, min(i + average_within + 1, n)):
                        other_scan = scans[j]
                        acc.append(other_scan)
                    if acc:
                        acq_info = scan.acquisition_information
                        scan = scan.average_with(
                            acc, dx=dx, num_threads=num_threads)
                        scan.acquisition_information = acq_info

            if average_across:
                pass

            if denoise > 0 and scan.arrays.mz.size > 0:
                scan = scan.denoise(denoise).transform(['savitsky_golay'])
            scan.pick_peaks(**kwargs)
            scan.peak_set.intensity_array = None
            scan.peak_set.mz_array = None
            drift_time = scan.drift_time
            for peak in scan:
                peak: FittedPeak
                if peak.intensity < minimum_intensity:
                    continue
                lff.handle_peak(peak, drift_time)

        del averager
        lff.split_sparse(max_gap_size, min_size).smooth_overlaps(
            error_tolerance)
        self.features = lff
        return self

    def deconvolute_features(self, averagine=None, scorer=None, truncate_after=0.95,
                             minimum_intensity=5, min_size=2, max_gap_size=0.25, copy=True,
                             **kwargs) -> 'IonMobilityFrame':
        from ms_deisotope.feature_map import feature_processor
        if averagine is None:
            from ms_deisotope.averagine import peptide
            averagine = peptide
        if scorer is None:
            from ms_deisotope.scoring import MSDeconVFitter
            scorer = MSDeconVFitter(1)
        if self.features is None:
            raise ValueError(
                "IM-MS Features must be extracted before they can be charge state deconvoluted")
        decon = feature_processor.LCMSFeatureProcessor(
            self.features, averagine, scorer, minimum_size=min_size,
            maximum_time_gap=max_gap_size, copy=copy)
        self.deconvoluted_features = decon.deconvolute(
            minimum_intensity=minimum_intensity, truncate_after=truncate_after, **kwargs)
        return self

    def pack(self, bind=False) -> 'ProcessedIonMobilityFrame':
        packed = ProcessedIonMobilityFrame(
            self.id, self.precursor_information, self.ms_level,
            self.time, self.index, self.features, self.deconvoluted_features,
            self.start_scan_index, self.end_scan_index,
            self.polarity, self.activation, self.acquisition_information,
            self.isolation_window, self.annotations,
            self.source if bind else None)
        return packed


class ProcessedIonMobilityFrame(FrameBase):

    id: str
    index: int


    def __init__(self, id, precursor_information, ms_level, time, index,
                 features=None, deconvoluted_features=None,
                 start_scan_index=None, end_scan_index=None,
                 polarity=None, activation=None, acquisition_information=None,
                 isolation_window=None, annotations=None, source=None):
        if annotations is None:
            annotations = dict()
        self.source = source

        self.id = id
        self.precursor_information = precursor_information
        self.ms_level = ms_level
        self.time = time
        self.index = index
        self.features = features
        self.deconvoluted_features = deconvoluted_features

        self.start_scan_index = start_scan_index
        self.end_scan_index = end_scan_index
        self.polarity = polarity
        self.activation = activation
        self.acquisition_information = acquisition_information
        self.isolation_window = isolation_window
        self.annotations = annotations


def weighted_centroid(feature):
    total = 0
    normalizer = 0
    for node in feature:
        weight = node.total_intensity()
        total += node.time * weight
        normalizer += weight
    return total / normalizer


def merge_envelopes(envelopes):
    base = envelopes[0].clone()
    for env in envelopes[1:]:
        for i, p in enumerate(env):
            base[i].intensity += p.intensity

    return base


def feature_to_peak(feature):
    peak_cluster = feature.peaks
    peak_cluster = [pi for p in peak_cluster for pi in p]
    total_intensity = sum(p.intensity for p in peak_cluster)
    mz = sum(p.mz * p.intensity for p in peak_cluster) / total_intensity
    neutral_mass = sum(
        p.neutral_mass * p.intensity for p in peak_cluster) / total_intensity
    most_abundant_mass = sum(
        p.most_abundant_mass * p.intensity for p in peak_cluster) / total_intensity
    a_to_a2_ratio = sum(
        p.a_to_a2_ratio * p.intensity for p in peak_cluster) / total_intensity
    average_mass = sum(
        p.average_mass * p.intensity for p in peak_cluster) / total_intensity
    signal_to_noise = sum(p.signal_to_noise *
                          p.intensity for p in peak_cluster) / total_intensity
    fwhm = sum(p.full_width_at_half_max *
               p.intensity for p in peak_cluster) / total_intensity
    area = sum(p.area * p.intensity for p in peak_cluster) / total_intensity
    score = sum(p.score * p.intensity for p in peak_cluster) / total_intensity
    charge = peak_cluster[0].charge
    envelope = merge_envelopes([p.envelope for p in peak_cluster])
    return IonMobilityDeconvolutedPeak(
        neutral_mass=neutral_mass, intensity=total_intensity, charge=charge, signal_to_noise=signal_to_noise,
        full_width_at_half_max=fwhm, index=-1, a_to_a2_ratio=a_to_a2_ratio, most_abundant_mass=most_abundant_mass,
        average_mass=average_mass, score=score, envelope=envelope, mz=mz, fit=None, chosen_for_msms=False,
        area=area, drift_time=weighted_centroid(feature))


def features_to_peak_set(features):
    peak_set = DeconvolutedPeakSet(tuple(map(feature_to_peak, features)))
    peak_set.reindex()
    return peak_set


class Generic3DIonMobilityFrameSource(IonMobilitySourceRandomAccessFrameSource):
    def __init__(self, loader, **kwargs):
        self.loader = loader
        self.initialize_frame_cache()
        self._producer = self._wrap_iterator(self.loader)

    def _frame_acquisition_information(self, data):
        return self.loader._acquisition_information(data)

    def _frame_activation(self, data):
        return self.loader._activation(data)

    def _frame_id(self, data):
        return self.loader._scan_id(data)

    def _frame_index(self, data):
        return self.loader._scan_index(data)

    def _frame_time(self, data):
        return self.loader._scan_time(data)

    def _frame_precursor_information(self, data):
        return self.loader._precursor_information(data)

    def _frame_ms_level(self, data):
        return self.loader._ms_level(data)

    def _frame_polarity(self, data):
        return self.loader._polarity(data)

    def _frame_isolation_window(self, data):
        return self.loader._isolation_window(data)

    def _frame_start_scan_index(self, data):
        return self._frame_index(data)

    def _frame_end_scan_index(self, data):
        return self._frame_index(data) + 1

    def _frame_arrays(self, data):
        from ms_deisotope.data_source.metadata.scan_traits import binary_data_arrays

        mz, intensity, data_arrays = self.loader._scan_arrays(data)
        ion_mobility = None
        ion_mobility_type = None
        for key, value in data_arrays.items():
            try:
                acc = key.accession
                term = binary_data_arrays[acc]
            except (KeyError, AttributeError):
                continue
            if term.is_a('ion mobility array'):
                ion_mobility = value
                ion_mobility_type = key
                break
        if ion_mobility_type is None:
            raise ValueError(
                "Expected an ion mobility array type, but none were found.")
        d = data_arrays.copy()
        d.pop(ion_mobility_type)
        array3d = RawDataArrays3D.from_arrays(
            mz, intensity, ion_mobility, ion_mobility_type, d)
        return array3d

    def _frame_scans(self, data):
        from ms_deisotope.data_source.metadata.scan_traits import ion_mobility_attribute

        arrays = self._frame_arrays(data)
        scan_id_base = self._frame_id(data)
        i = 1
        scans = []
        for drift_time, arrays_i in arrays.unstack():
            subdat = data.copy()
            subdat['m/z array'] = arrays_i.mz
            subdat['intensity array'] = arrays_i.intensity
            subdat['id'] = scan_id_base + (' scan=%d' % i)
            i += 1
            scan = self.loader._make_scan(subdat)
            acq = scan.acquisition_information
            event = acq[0]
            tp = event.ion_mobility_type
            if tp:
                event._ion_mobility.remove_ion_mobility_type(tp)
            event._ion_mobility.add_ion_mobility(
                ion_mobility_attribute, drift_time)
            scans.append(scan)
        return scans

    def _frame_ion_mobilities(self, data):
        return self._frame_arrays(data).distinct_ion_mobility

    def get_frame_by_index(self, index: int) -> IonMobilityFrame:
        scan = self.loader.get_scan_by_index(index)
        frame = self._make_frame(scan._data)
        self._cache_frame(frame)
        return frame

    def get_frame_by_id(self, id: str) -> IonMobilityFrame:
        scan = self.loader.get_scan_by_id(id)
        frame = self._make_frame(scan._data)
        self._cache_frame(frame)
        return frame

    def get_frame_by_time(self, time: float) -> IonMobilityFrame:
        scan = self.loader.get_scan_by_time(time)
        frame = self._make_frame(scan._data)
        self._cache_frame(frame)
        return frame

    def __len__(self):
        return len(self.loader)

    def __getitem__(self, i):
        result = self.loader[i]
        if isinstance(result, list):
            out = []
            for scan in result:
                frame = self._make_frame(scan._data)
                self._cache_frame(frame)
                out.append(frame)
            return out
        else:
            scan = result
            frame = self._make_frame(scan._data)
            self._cache_frame(frame)
            return frame

    def _wrap_iterator(self, iterator):
        for val in iterator:
            if isinstance(val, ScanBunch):
                if val.precursor is None:
                    precursor = None
                else:
                    precursor = self._make_frame(val.precursor._data)
                    self._cache_frame(precursor)
                products = []
                for product in val.products:
                    product = self._make_frame(product._data)
                    self._cache_frame(product)
                    products.append(product)

                yield ScanBunch(precursor, products)
            else:
                frame = self._make_frame(val._data)
                self._cache_frame(frame)
                yield frame

    def _default_frame_iterator(self, start_index=None):
        self.loader.start_from_scan(index=start_index, grouped=False)
        return self._wrap_iterator(self.loader)

    def make_frame_iterator(self, iterator=None, grouped=False, **kwargs):
        self.loader.make_iterator(iterator, grouped=grouped, **kwargs)
        return self._wrap_iterator(self.loader)

    def start_from_frame(self, scan_id=None, rt=None, index=None, require_ms1=True, grouped=True, **kwargs):
        self.loader.start_from_scan(
            scan_id=scan_id, rt=rt, index=index, require_ms1=require_ms1, grouped=grouped, **kwargs)
        self._producer = self._wrap_iterator(self.loader)
        return self

    def __iter__(self):
        return self

    def __next__(self):
        return next(self._producer)

    def next(self):
        return next(self._producer)

    def reset(self):
        self.loader.reset()
        self.initialize_frame_cache()
        return self


class FramedIonMobilityFrameSource(IonMobilitySourceRandomAccessFrameSource):
    def __init__(self, loader, **kwargs):
        self.loader = loader
        self.initialize_frame_cache()
        self.scan_index_range_to_frame = dict()

    def _find_cycle_start(self, scan):
        dt = scan.drift_time
        if dt is None:
            return scan
        index = scan.index
        best_scan = scan
        while index >= 0:
            prev = self.loader[index]
            p_dt = prev.drift_time
            if p_dt is None:
                return best_scan
            if p_dt < dt:
                best_scan = scan
            else:
                return best_scan
            index -= 1
        return best_scan

    def _collect_cycle(self, scan):
        acc = []
        start_scan = self._find_cycle_start(scan)
        dt = start_scan.drift_time
        if dt is None:
            raise ValueError("Could not find a start scan with drift time from %r" % (scan, ))
        acc.append(start_scan)
        index = start_scan.index + 1
        while True:
            next_scan = self.loader.get_scan_by_index(index)
            next_dt = next_scan.drift_time
            if next_dt is None:
                break
            if next_dt < dt:
                break
            dt = next_dt
            acc.append(next_scan)
            index += 1
        return acc

    def _frame_id(self, frame):
        scans = self._frame_scans(frame)
        return ','.join([s.id for s in scans])

    def _frame_acquisition_information(self, data):
        return self.loader._acquisition_information(data)

    def _frame_activation(self, data):
        scans = self._frame_scans(data)
        return scans[0].activation

    def _frame_index(self, data):
        scans = self._frame_scans(data)
        return scans[0].index

    def _frame_time(self, data):
        scans = self._frame_scans(data)
        return scans[0].scan_time

    def _frame_ms_level(self, data):
        scans = self._frame_scans(data)
        return scans[0].ms_level

    def _frame_polarity(self, data):
        scans = self._frame_scans(data)
        return scans[0].polarity

    def _frame_isolation_window(self, data):
        scans = self._frame_scans(data)
        return scans[0].isolation_window

    def _frame_start_scan_index(self, data):
        scans = self._frame_scans(data)
        return scans[0].index

    def _frame_end_scan_index(self, data):
        scans = self._frame_scans(data)
        return scans[-1].index + 1

    def _frame_scans(self, data):
        try:
            return data['scans']
        except KeyError:
            pass
        source_id = data['source_scan_id']
        scan = self.loader.get_scan_by_id(source_id)
        scans = self._collect_cycle(scan)
        data['scans'] = scans
        data['scan_index_range'] = (scans[0].index, scans[-1].index + 1)
        return data['scans']

    def _make_frame_from_scan(self, scan):
        return {
            "source_scan_id": scan.id,
            "source_scan_index": scan.index,
        }

    def get_scan_by_index(self, index):
        return self.loader.get_scan_by_index(index)
